# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight convenience container for various model implementations."""

import collections
import functools
from typing import Any, Callable, Dict, Optional, Mapping, NamedTuple, Tuple

import dataclasses
from fedjax.core import metrics
from fedjax.core.typing import Batch
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey
from fedjax.core.typing import Updates
import haiku as hk
import immutabledict
import jax
import jax.numpy as jnp

MetricsFn = Callable[[Batch, jnp.ndarray], metrics.Metric]


class BackwardPassOutput(NamedTuple):
  """Struct for backward pass output that can be passed to JAX transformations.

  Attributes:
    grads: Gradients of same structure as given params.
    loss: Scalar loss for given batch.
    num_examples: Number of examples seen in a given batch.
  """
  grads: Updates
  loss: jnp.ndarray
  num_examples: float


@dataclasses.dataclass(frozen=True)
class Model:
  """Cotainer class for models.

  Works for Haiku (go/dm-haiku) and jax.experimental.stax.
  NOTE: Users should not create this directly but use `create_model_from_haiku`
  or `create_model_from_stax` instead.

  Attributes:
    init_fn: Model parameter initialization function.
    apply_fn: Function that produces model predictions for input batch.
    loss_fn: Loss function that takes input batch and model predictions and
      returns scalar loss.
    reg_fn: Regularization function that takes parameters in and returns
      a scalar regularizer value.
    metrics_fn_map: Ordered mapping of metric names to metric functions that
      take input batch and model predictions and return metric values.
    train_kwargs: Keyword arguments passed to apply for training.
    test_kwargs: Keyword arguments passed to apply for testing.
    modify_grads_fn: Callable that modifies input gradients.
  """
  init_fn: Callable[[PRNGKey], Params]
  apply_fn: Callable[..., jnp.ndarray]
  loss_fn: MetricsFn
  reg_fn: Callable[[Params], jnp.ndarray] = lambda p: 0.
  metrics_fn_map: Mapping[str, MetricsFn] = immutabledict.immutabledict()
  train_kwargs: Mapping[str, Any] = immutabledict.immutabledict()
  test_kwargs: Mapping[str, Any] = immutabledict.immutabledict()
  modify_grads_fn: Callable[[Updates], Updates] = lambda g: g

  def init_params(self, rng: PRNGKey) -> Params:
    return self.init_fn(rng)

  @functools.partial(jax.jit, static_argnums=0)
  def backward_pass(self, params: Params, batch: Batch,
                    rng: Optional[PRNGKey]) -> BackwardPassOutput:
    """Runs backward pass and returns BackwardPassOutput."""

    def loss_fn(p):
      preds = self.apply_fn(p, rng, batch, **self.train_kwargs)
      loss_metric = self.loss_fn(batch, preds)
      loss = loss_metric.result() + self.reg_fn(p)
      return loss, loss_metric.count

    grad_fn = jax.value_and_grad(loss_fn, has_aux=True)
    (loss, num_examples), grads = grad_fn(params)
    grads = self.modify_grads_fn(grads)
    return BackwardPassOutput(grads=grads, loss=loss, num_examples=num_examples)

  @functools.partial(jax.jit, static_argnums=0)
  def evaluate(self, params: Params, batch: Batch) -> Dict[str, metrics.Metric]:
    """Evaluates model on input batch."""
    rng = None
    preds = self.apply_fn(params, rng, batch, **self.test_kwargs)
    loss_metric = self.loss_fn(batch, preds)
    num_examples = loss_metric.count
    metrics_dict = collections.OrderedDict(
        loss=loss_metric,
        regularizer=metrics.MeanMetric(
            total=self.reg_fn(params), count=num_examples),
        num_examples=metrics.CountMetric(count=num_examples))
    for metric_name, metric_fn in self.metrics_fn_map.items():
      metrics_dict[metric_name] = metric_fn(batch, preds)
    return metrics_dict


def _get_defaults(reg_fn, metrics_fn_map, train_kwargs, test_kwargs):
  """Builds default values."""
  if reg_fn is None:
    reg_fn = lambda p: 0.
  metrics_fn_map = metrics_fn_map or collections.OrderedDict()
  train_kwargs = train_kwargs or {}
  test_kwargs = test_kwargs or {}
  frozen_metrics_fn_map = immutabledict.immutabledict(metrics_fn_map)
  frozen_train_kwargs = immutabledict.immutabledict(train_kwargs)
  frozen_test_kwargs = immutabledict.immutabledict(test_kwargs)
  return reg_fn, frozen_metrics_fn_map, frozen_train_kwargs, frozen_test_kwargs


def create_model_from_haiku(
    transformed_forward_pass: hk.Transformed,
    sample_batch: Batch,
    loss_fn: MetricsFn,
    reg_fn: Optional[Callable[[Params], jnp.ndarray]] = None,
    metrics_fn_map: Optional[Mapping[str, MetricsFn]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    test_kwargs: Optional[Mapping[str, Any]] = None,
    non_trainable_module_names: Tuple[str] = ()) -> Model:  # pytype: disable=annotation-type-mismatch
  """Creates Model after applying defaults and haiku specific preprocessing.

  Args:
    transformed_forward_pass: Transformed forward pass from `hk.transform`.
    sample_batch: Example input batch used to determine model parameter shapes.
    loss_fn: Loss function that takes input batch and model predictions and
      returns scalar loss.
    reg_fn: Regularization function that takes parameters in and returns
      a scalar regularizer value. Defaults to no regularization.
    metrics_fn_map: Ordered mapping of metric names to metric functions that
      take input batch and model predictions and return metric values that will
      be freezed for immutability. Defaults to empty frozen dictionary.
    train_kwargs: Keyword arguments passed to model for training that will be
      freezed for immutability. Defaults to empty frozen dictionary.
    test_kwargs: Keyword arguments passed to model for testing that will be
      freezed for immutability. Defaults to empty frozen dictionary.
    non_trainable_module_names: List of `hk.Module` names whose parameters
      should not to be updated during training.

  Returns:
    Model
  """
  reg_fn, metrics_fn_map, train_kwargs, test_kwargs = _get_defaults(
      reg_fn, metrics_fn_map, train_kwargs, test_kwargs)

  def ignore_grads(grads):
    predicate = lambda module_name, *_: module_name in non_trainable_module_names
    non_trainable, trainable = hk.data_structures.partition(predicate, grads)
    non_trainable = jax.tree_map(jnp.zeros_like, non_trainable)
    return hk.data_structures.merge(non_trainable, trainable)

  def init(rng):
    return transformed_forward_pass.init(rng, sample_batch)

  return Model(init, transformed_forward_pass.apply, loss_fn, reg_fn,
               metrics_fn_map, train_kwargs, test_kwargs, ignore_grads)


def create_model_from_stax(
    stax_init_fn: Callable[..., Params],
    stax_apply_fn: Callable[..., jnp.ndarray],
    sample_shape: Tuple[int, ...],
    loss_fn: MetricsFn,
    input_key: str = 'x',
    reg_fn: Optional[Callable[[Params], jnp.ndarray]] = None,
    metrics_fn_map: Optional[Mapping[str, MetricsFn]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    test_kwargs: Optional[Mapping[str, Any]] = None) -> Model:
  """Creates Model after applying defaults and stax specific preprocessing.

  Args:
    stax_init_fn: Initialization function returned from `stax.serial`.
    stax_apply_fn: Model forward_pass pass function returned from `stax.serial`.
    sample_shape: The expected shape of the input to the model.
    loss_fn: Loss function that takes input batch and model predictions and
      returns scalar loss.
    input_key: Key name for the input from batch mapping.
    reg_fn: Regularization function that takes parameters in and returns
      a scalar regularizer value. Defaults to no regularization.
    metrics_fn_map: Ordered mapping of metric names to metric functions that
      take input batch and model predictions and return metric values that will
      be freezed for immutability. Defaults to empty frozen dictionary.
    train_kwargs: Keyword arguments passed to model for training that will be
      freezed for immutability. Defaults to empty frozen dictionary.
    test_kwargs: Keyword arguments passed to model for testing that will be
      freezed for immutability. Defaults to empty frozen dictionary.

  Returns:
    Model
  """
  reg_fn, metrics_fn_map, train_kwargs, test_kwargs = _get_defaults(
      reg_fn, metrics_fn_map, train_kwargs, test_kwargs)

  def init(rng):
    _, params = stax_init_fn(rng, sample_shape)
    return params

  def apply(params, rng, batch, **kwargs):
    return stax_apply_fn(params, batch[input_key], rng=rng, **kwargs)

  return Model(init, apply, loss_fn, reg_fn, metrics_fn_map, train_kwargs,
               test_kwargs)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for working with tff.simulation.ClientData."""

import functools
from typing import Any, Iterable, Iterator, List, NamedTuple, Optional, Union

from fedjax.core.typing import FederatedData
import tensorflow as tf


def create_tf_dataset_for_clients(
    federated_data: FederatedData,
    client_ids: Optional[List[str]] = None) -> tf.data.Dataset:
  """Creates a single dataset by combining multiple input client datasets.

  Args:
    federated_data: Federated dataset.
    client_ids: Clients to combine. If None, combines all in client_data.

  Returns:
    A tf.data.Dataset of ordered mapping of combined examples.
  """
  client_ids = client_ids or federated_data.client_ids

  # create_tf_dataset_from_all_clients has fewer file reads than per client.
  if sorted(client_ids) == sorted(federated_data.client_ids):
    return federated_data.create_tf_dataset_from_all_clients()

  client_datasets = [
      federated_data.create_tf_dataset_for_client(c) for c in client_ids
  ]
  # Concatenate multiple client tf.data.Datasets into one tf.data.Dataset.
  return functools.reduce(lambda a, b: a.concatenate(b), client_datasets)


class ClientDataHParams(NamedTuple):
  """Hyperparameters for client data preparation.

  Attributes:
    batch_size: Batch size for training single or multiple clients.
    num_epochs: Number of epochs over data of single or multiple clients.
    drop_remainder: Whether to drop the last batch if it's < batch_size.
    shuffle_buffer_size: Maximum number of elements that will be buffered when
      prefetching. If 0, don't shuffle.
    num_batches: Maximum number of batches to include. Defaults to all.
  """
  batch_size: int = 1
  num_epochs: int = 1
  drop_remainder: bool = False
  shuffle_buffer_size: int = 0
  num_batches: int = -1


@tf.function
def preprocess_tf_dataset(dataset: tf.data.Dataset,
                          hparams: ClientDataHParams) -> tf.data.Dataset:
  """Preprocesses dataset according to the dataset hyperparmeters.

  Args:
    dataset: Dataset with a mapping element structure.
    hparams: Hyper parameters for dataset preparation.

  Returns:
    Preprocessed dataset.
  """
  dataset = dataset.repeat(hparams.num_epochs)
  if hparams.shuffle_buffer_size:
    dataset = dataset.shuffle(hparams.shuffle_buffer_size)
  dataset = (dataset.batch(hparams.batch_size,
                           drop_remainder=hparams.drop_remainder)
             .prefetch(1))
  return dataset.take(hparams.num_batches)


DatasetOrIterable = Union[tf.data.Dataset, Iterable[Any]]


def iterate(dataset: DatasetOrIterable) -> Iterator[Any]:
  """Unified iteration over TF dataset or JAX supported types.

  This allows functions to take an argument from a range of types, while
  iterating over them in the same way. Currently the following are supported,

  - tf.data.Dataset: Iteration over a tf.data.Dataset directly produces numpy
  arrays.

  - Any other iterable: Such an iterable is directly iterated over.


  For example:

  def f(xs: DatasetOrIterable):
    total = 0
    for x in iterate(xs):
      total += x
    return total

  # Works with TF dataset.
  f(tf.data.Dataset.range(4))
  # Works with a numpy array list.
  f([np.ones((2, 2)), np.ones((1, 2)), np.ones((2, 1))])

  Args:
    dataset: Object to iterate over.

  Returns:
    An iterator.
  """
  if isinstance(dataset, tf.data.Dataset):
    return dataset.as_numpy_iterator()
  return iter(dataset)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.training.checkpoint."""

import os

from fedjax.core import test_util
from fedjax.training import checkpoint
import tensorflow as tf


class CheckpointTest(tf.test.TestCase):

  def setUp(self):
    super().setUp()
    self._root_dir = self.get_temp_dir()

  def _write_file(self, basename):
    path = os.path.join(self._root_dir, basename)
    with tf.io.gfile.GFile(path, mode='w') as f:
      f.write(basename)

  def test_save_checkpoint(self):
    checkpoint.save_checkpoint(self._root_dir, test_util.create_mock_state())

    self.assertEqual(
        tf.io.gfile.listdir(self._root_dir), ['checkpoint_00000000'])

  def test_save_checkpoint_keep(self):
    state = test_util.create_mock_state()

    for i in range(3):
      checkpoint.save_checkpoint(self._root_dir, state, round_num=i, keep=2)

    self.assertCountEqual(
        tf.io.gfile.listdir(self._root_dir),
        ['checkpoint_00000001', 'checkpoint_00000002'])

  def test_get_checkpoint_paths(self):
    self._write_file('checkpoint_00000100')
    self._write_file('checkpoint_00000000')
    self._write_file('checkpoint_00000003')
    # Similarly named files that shouldn't be counted as valid checkpoints.
    self._write_file('checkpoint_haha')
    self._write_file('checkpoint_0ha')
    self._write_file('checkpoint_00000004.tf.ckpt')

    checkpoint_paths = checkpoint._get_checkpoint_paths(
        os.path.join(self._root_dir, 'checkpoint_'))

    basenames = [os.path.basename(cp) for cp in checkpoint_paths]
    self.assertEqual(
        basenames,
        ['checkpoint_00000000', 'checkpoint_00000003', 'checkpoint_00000100'])

  def test_load_latest_checkpoint(self):
    state_1 = test_util.create_mock_state(seed=1)
    state_2 = test_util.create_mock_state(seed=2)
    checkpoint.save_checkpoint(self._root_dir, state_1, round_num=10, keep=2)
    checkpoint.save_checkpoint(self._root_dir, state_2, round_num=3, keep=2)

    latest_state, latest_round_num = checkpoint.load_latest_checkpoint(
        root_dir=self._root_dir)

    self.assertEqual(latest_round_num, 10)
    tf.nest.map_structure(self.assertAllEqual, latest_state, state_1)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Lightweight convenience container for various model implementations."""

import functools
from typing import Any, Callable, Dict, Iterable, Optional, Mapping, Tuple

from fedjax import core
from fedjax.experimental import metrics
from fedjax.experimental.typing import BatchExample
from fedjax.experimental.typing import BatchPrediction

import haiku as hk
import immutabledict
import jax
import jax.numpy as jnp

Params = core.Params
PRNGKey = core.PRNGKey

# Typically unnormalized model forward pass output.
BatchTrainOutput = jnp.ndarray
BatchEvalPrediction = BatchPrediction


@core.dataclass
class Model:
  """Container class for models.

  `Model` exists to provide easy access to predefined neural network models.
  It is meant to contain all the information needed for standard centralized
  training and evaluation. Non-standard training methods can be built upon the
  information avaiable in `Model` along with any additional information
  (e.g. interpolation can be implemented as a composition of two models along
  with an interpolation weight).

  Works for Haiku (go/dm-haiku) and jax.experimental.stax. We strongly recommend
  using the `Model.new()` factory method to construct `Model` objects.

  The expected usage of `Model` is as follows:

  ```
  # Training.
  step_size = 0.1
  rng = jax.random.PRNGKey(0)
  params = model.init(rng)

  def loss(params, batch, rng):
    preds = model.apply_for_train(params, batch, rng)
    return jnp.sum(model.train_loss(batch, preds))

  grad_fn = jax.grad(loss)
  for batch in batches:
    rng, use_rng = jax.random.split(rng)
    grads = grad_fn(params, batch, use_rng)
    params = jax.tree_util.tree_multimap(lambda a, b: a - step_size * b,
                                         params, grads)

  # Evaluation.
  print(fedjax.experimental.model.evaluate_model(model, params, batches))
  # Example output:
  # {'loss': 2.3, 'accuracy': 0.2}
  ```

  The following is an example using `Model` compositionally as a building block
  to impelement model interpolation.

  ```
  def interpolate(model_1, model_2, init_weight):

    @jax.jit
    def init(rng):
      rng_1, rng_2 = jax.random.split(rng)
      params_1 = model_1.init(rng_1)
      params_2 = model_2.init(rng_2)
      return params_1, params_2, init_weight

    @jax.jit
    def apply_for_train(params, input, rng):
      rng_1, rng_2 = jax.random.split(rng)
      params_1, params_2, weight = params
      return (model_1.apply_for_train(params_1, input, rng_1) * weight +
              model_2.apply_for_train(params_1, input, rng_2) * (1 - weight))

    @jax.jit
    def apply_for_eval(params, input):
      params_1, params_2, weight = params
      return (model_1.apply_for_eval(params_1, input) * weight +
              model_2.apply_for_eval(params_2, input) * (1 - weight))

    return fedjax.experimental.model.Model.new(init,
                                               apply_for_train,
                                               apply_for_eval,
                                               model_1.train_loss,
                                               model_1.eval_metrics)

  model = interpolate(model_1, model_2, init_weight=0.5)
  ```

  Attributes:
    init: Initialization function that takes a seed `PRNGKey` and returns
      a PyTree of initialized parameters (i.e. model weights). These parameters
      will be passed as input into `apply_for_train` and `apply_for_eval`.
      Any trainable weights for a model that are modified in the training loop
      should be contained inside of these parameters.
    apply_for_train: Function that takes the parameters PyTree, batch of
      examples, and `PRNGKey` as inputs and outputs the model predictions for
      training that are then passed into `train_loss`. This considers strategies
      such as dropout.
    apply_for_eval: Function that usually takes the parameters PyTree and batch
      of examples as inputs and outputs the model predictions for evaluation
      that are then passed into the metric functions in `eval_metric_fns`. This
      is defined separately from `apply_for_train` to avoid having to specify
      inputs like `PRNGKey` that are not used in evaluation.
    train_loss: Loss function for training that takes batch of examples and
      model output from `apply_for_train` as input that outputs per example
      loss. This will typically called inside a `jax.grad` wrapped function to
      compute gradients.
    eval_metrics: Ordered mapping of evaluation metric names to `Metric`s. These
      `Metric`s are defined for single examples and will be consumed in
      `evaluate_model`.
  """
  init: Callable[[PRNGKey], Params]
  apply_for_train: Callable[[Params, BatchExample, Optional[PRNGKey]],
                            BatchTrainOutput]
  apply_for_eval: Callable[[Params, BatchExample], BatchEvalPrediction]
  train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray]
  eval_metrics: Mapping[str, metrics.Metric]

  @classmethod
  def new(cls, init: Callable[[PRNGKey], Params],
          apply_for_train: Callable[[Params, BatchExample, Optional[PRNGKey]],
                                    BatchTrainOutput],
          apply_for_eval: Callable[[Params, BatchExample], BatchEvalPrediction],
          train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray],
          eval_metrics: Mapping[str, metrics.Metric]) -> 'Model':
    """Freezes mutable arguments to `Model` to make it `jax.jit` friendly."""
    return cls(init, apply_for_train, apply_for_eval, train_loss,
               immutabledict.immutabledict(eval_metrics))


def create_model_from_haiku(
    transformed_forward_pass: hk.Transformed,
    sample_batch: BatchExample,
    train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray],
    eval_metrics: Optional[Mapping[str, metrics.Metric]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    eval_kwargs: Optional[Mapping[str, Any]] = None) -> Model:
  """Creates Model after applying defaults and haiku specific preprocessing.

  Args:
    transformed_forward_pass: Transformed forward pass from `hk.transform`.
    sample_batch: Example input used to determine model parameter shapes.
    train_loss: Loss function for training that outputs per example loss.
    eval_metrics: Mapping of evaluation metric names to `Metric`s. These
      `Metric`s are defined for single examples and will be consumed in
      `evaluate_model`.
    train_kwargs: Keyword arguments passed to model for training.
    eval_kwargs: Keyword arguments passed to model for evaluation.

  Returns:
    Model
  """
  eval_metrics = eval_metrics or {}
  train_kwargs = train_kwargs or {}
  eval_kwargs = eval_kwargs or {}

  @jax.jit
  def init(rng):
    return transformed_forward_pass.init(rng, sample_batch)

  @jax.jit
  def apply_for_train(params, batch, rng=None):
    return transformed_forward_pass.apply(params, rng, batch, **train_kwargs)

  @jax.jit
  def apply_for_eval(params, batch):
    return transformed_forward_pass.apply(params, None, batch, **eval_kwargs)

  return Model.new(init, apply_for_train, apply_for_eval, train_loss,
                   eval_metrics)


def create_model_from_stax(
    stax_init: Callable[..., Params],
    stax_apply: Callable[..., jnp.ndarray],
    sample_shape: Tuple[int, ...],
    train_loss: Callable[[BatchExample, BatchTrainOutput], jnp.ndarray],
    eval_metrics: Optional[Mapping[str, metrics.Metric]] = None,
    train_kwargs: Optional[Mapping[str, Any]] = None,
    eval_kwargs: Optional[Mapping[str, Any]] = None,
    input_key: str = 'x') -> Model:
  """Creates Model after applying defaults and stax specific preprocessing.

  Args:
    stax_init: Initialization function returned from `stax.serial`.
    stax_apply: Model forward_pass pass function returned from `stax.serial`.
    sample_shape: The expected shape of the input to the model.
    train_loss: Loss function for training that outputs per example loss.
    eval_metrics: Mapping of evaluation metric names to `Metric`s. These
      `Metric`s are defined for single examples and will be consumed in
      `evaluate_model`.
    train_kwargs: Keyword arguments passed to model for training.
    eval_kwargs: Keyword arguments passed to model for evaluation.
    input_key: Key name for the input in batch mapping.

  Returns:
    Model
  """
  eval_metrics = eval_metrics or {}
  train_kwargs = train_kwargs or {}
  eval_kwargs = eval_kwargs or {}

  @jax.jit
  def init(rng):
    _, params = stax_init(rng, sample_shape)
    return params

  @jax.jit
  def apply_for_train(params, batch, rng=None):
    return stax_apply(params, batch[input_key], rng=rng, **train_kwargs)

  @jax.jit
  def apply_for_eval(params, batch):
    return stax_apply(params, batch[input_key], **eval_kwargs)

  return Model.new(init, apply_for_train, apply_for_eval, train_loss,
                   eval_metrics)


@functools.partial(jax.jit, static_argnums=(0, 1))
def _evaluate_model_step(mask_key: str, model: Model, params: Params,
                         batch: BatchExample,
                         stat: metrics.Stat) -> Dict[str, metrics.Stat]:
  """Evaluates model for one batch and returns merged `Stat`.

  Args:
    mask_key: Reserved key name in example mapping for the example level mask.
    model: `Model` container with `apply_for_eval` and `eval_metrics`.
    params: Pytree of model parameters to be evaluated.
    batch: Batch of N examples.
    stat: Intermediate `Stat` from the previous step to be accumulated in the
      current step.

  Returns:
    A dictionary of intermediate evaluation `Stat`s.
  """
  try:
    mask = batch[mask_key].astype(jnp.bool_)
  except KeyError:
    mask = jnp.ones([len(next(iter(batch.values())))], dtype=jnp.bool_)
  pred = model.apply_for_eval(params, batch)
  new_stat = {
      k: metrics.evaluate_batch(metric, batch, pred, mask)
      for k, metric in model.eval_metrics.items()
  }
  return jax.tree_util.tree_multimap(
      lambda a, b: a.merge(b),
      stat,
      new_stat,
      is_leaf=lambda v: isinstance(v, metrics.Stat))


def evaluate_model(model: Model,
                   params: Params,
                   batches: Iterable[BatchExample],
                   mask_key='mask') -> Dict[str, jnp.ndarray]:
  """Evaluates model for multiple batches and returns final results.

  This is the recommended way to compute evaluation metrics for a given model.

  Args:
    model: `Model` container with `apply_for_eval` and `eval_metrics`.
    params: Pytree of model parameters to be evaluated.
    batches: Multiple batches to compute and aggregate evaluation metrics over.
    mask_key: Reserved key name in example mapping for the example level mask.

  Returns:
    A dictionary of evaluation `Metric` results.
  """
  stat = {k: metric.zero() for k, metric in model.eval_metrics.items()}
  for batch in batches:
    stat = _evaluate_model_step(mask_key, model, params, batch, stat)
  return jax.tree_util.tree_map(
      lambda x: x.result(), stat, is_leaf=lambda v: isinstance(v, metrics.Stat))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""FedJAX types."""

from typing import Any, Dict, Mapping, Union
import haiku as hk
from jax.experimental import optimizers
import jax.numpy as jnp
import numpy as np
import optax
import tensorflow_federated as tff

# Flexible Params definition to support haiku as well as stax.
Params = Any
Updates = Params
# Flexible OptState definition to support optax and jax.experimental.optimizers.
OptState = Union[optax.OptState, optimizers.OptimizerState]

# Mapping of feature names to feature values.
Batch = Mapping[str, np.ndarray]
# Mapping of metric names to metric results.
MetricResults = Dict[str, jnp.ndarray]

# Missing JAX types.
PRNGKey = jnp.ndarray
# TODO(b/162863493): Replace with stricter definition if possible.
PyTree = Any

# Convenient FedJAX typecasts.
FederatedData = tff.simulation.ClientData
PRNGSequence = hk.PRNGSequence



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Agnostic federated averaging implementation.

Based on the paper:

Agnostic Federated Learning
    Mehryar Mohri, Gary Sivek, Ananda Theertha Suresh
    https://arxiv.org/abs/1902.00146

This implementation only supports the horizontal partitions case in which each
client belongs to only a single domain (e.g. user locale).
"""

import enum
from typing import Callable, List, NamedTuple, Tuple

from fedjax import core
import jax.numpy as jnp


class AgnosticFedAvgState(NamedTuple):
  """State of server for AgnosticFedAvg passed between rounds.

  Attributes:
    params: A pytree representing the server model parameters.
    server_opt_state: A pytree representing the server optimizer state.
    domain_weights: Weights per domain applied to weight in weighted average.
  """
  params: core.Params
  server_opt_state: core.OptState
  domain_weights: jnp.ndarray


# TODO(b/162840323): Implement SGD.
@enum.unique
class DomainAlgorithm(enum.Enum):
  """Domain weight update algorithm options.

  Attributes:
    EG: Exponentiated gradient update.
    NONE: No update.
  """
  EG = 'EG'
  NONE = 'NONE'


class AgnosticFedAvgHParams(NamedTuple):
  """Hyperparameters for agnostic federated averaging.

  Attributes:
    train_data_hparams: Hyperparameters for training client data preparation.
    init_domain_weights: Initial weights per domain that sums to 1.
    domain_id_fn: Given client id returns the domain id for that client.
    domain_learning_rate: Learning rate for domain weight update.
    domain_algorithm: Algorithm used to update domain weights each round.
  """
  train_data_hparams: core.ClientDataHParams
  init_domain_weights: Tuple[float, ...]
  domain_id_fn: Callable[[str], int] = lambda _: 0
  domain_learning_rate: float = 0.1
  domain_algorithm: DomainAlgorithm = DomainAlgorithm.EG


def _update_domain_weights(domain_weights: jnp.ndarray,
                           domain_loss: jnp.ndarray,
                           domain_learning_rate: float,
                           domain_algorithm: DomainAlgorithm) -> jnp.ndarray:
  """Updates domain weights following input algorithm.

  Args:
    domain_weights: Weights per domain that sums to 1.
    domain_loss: Average loss per domain.
    domain_learning_rate: Learning rate for domain update.
    domain_algorithm: Domain weight update strategy.

  Returns:
    Updated domain weights.

  Raises:
    ValueError: If an unsupported DomainAlgorithm is specified.
  """
  if domain_algorithm == DomainAlgorithm.EG:
    new_domain_weights = domain_weights * jnp.exp(
        domain_learning_rate * domain_loss)
    new_domain_weights = jnp.maximum(new_domain_weights,
                                     jnp.zeros_like(new_domain_weights))
    return new_domain_weights / jnp.sum(new_domain_weights)
  elif domain_algorithm == DomainAlgorithm.NONE:
    return domain_weights
  else:
    raise ValueError(f'Unsupported domain algorithm {domain_algorithm}.')


class AgnosticFedAvg(core.FederatedAlgorithm):
  """Agnostic federated averaging algorithm."""

  def __init__(self,
               federated_data: core.FederatedData,
               model: core.Model,
               client_optimizer: core.Optimizer,
               server_optimizer: core.Optimizer,
               hparams: AgnosticFedAvgHParams,
               rng_seq: core.PRNGSequence):
    """Initializes HypCluster algorithm.

    Args:
      federated_data: Federated data separated per client.
      model: Model implementation.
      client_optimizer: Client optimizer.
      server_optimizer: Server optimizer.
      hparams: Hyperparameters for agnostic federated averaging.
      rng_seq: Iterator of JAX random keys.
    """
    self._federated_data = federated_data
    self._model = model
    self._client_optimizer = client_optimizer
    self._server_optimizer = server_optimizer
    self._hparams = hparams
    self._rng_seq = rng_seq
    self._client_trainer = core.DefaultClientTrainer(model, client_optimizer)

  @property
  def federated_data(self) -> core.FederatedData:
    return self._federated_data

  @property
  def model(self) -> core.Model:
    return self._model

  def init_state(self) -> AgnosticFedAvgState:
    params = self._model.init_params(next(self._rng_seq))
    server_opt_state = self._server_optimizer.init_fn(params)
    domain_weights = jnp.array(self._hparams.init_domain_weights)
    return AgnosticFedAvgState(
        params=params,
        server_opt_state=server_opt_state,
        domain_weights=domain_weights)

  def run_round(self, state: AgnosticFedAvgState,
                client_ids: List[str]) -> AgnosticFedAvgState:
    """Runs one round of agnostic federated averaging."""
    client_metrics = core.evaluate_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        model=self.model,
        params=state.params,
        client_data_hparams=self._hparams.train_data_hparams)

    domain_ids = []
    num_domains = len(state.domain_weights)
    domain_loss = [0.0] * num_domains
    domain_num = [0.0] * num_domains
    for client_id, metrics in zip(client_ids, client_metrics):
      domain_id = self._hparams.domain_id_fn(client_id)
      domain_ids.append(domain_id)
      if metrics:
        domain_loss[domain_id] += metrics['loss'] * metrics['num_examples']
        domain_num[domain_id] += metrics['num_examples']
    domain_loss = jnp.array(domain_loss)
    domain_num = jnp.array(domain_num)

    # Train model per client.
    client_states = core.train_multiple_clients(
        federated_data=self.federated_data,
        client_ids=client_ids,
        client_trainer=self._client_trainer,
        init_client_trainer_state=self._client_trainer.init_state(state.params),
        rng_seq=self._rng_seq,
        client_data_hparams=self._hparams.train_data_hparams)

    # Scale domain weights by the proportion of total domain examples.
    scaled_dw = state.domain_weights / domain_num
    scaled_dw = jnp.where(domain_num == 0., 0., scaled_dw)

    # Weighted average of param delta across clients.
    def select_delta_params_and_weight(element):
      idx, client_state = element
      delta_params = core.tree_multimap(lambda a, b: a - b, state.params,
                                        client_state.params)
      # Use client weight scaled by domain weight weight for weighted average.
      domain_id = domain_ids[idx]
      client_weight = scaled_dw[domain_id] * client_state.num_examples
      return delta_params, client_weight

    delta_params_and_weight = map(select_delta_params_and_weight,
                                  enumerate(client_states))
    delta_params = core.tree_mean(delta_params_and_weight)

    # Server update.
    avg_loss_per_domain = jnp.where(domain_num == 0., 0.,
                                    domain_loss / domain_num)
    domain_weights = _update_domain_weights(
        domain_weights=state.domain_weights,
        domain_loss=avg_loss_per_domain,
        domain_learning_rate=self._hparams.domain_learning_rate,
        domain_algorithm=self._hparams.domain_algorithm)
    updates, server_opt_state = self._server_optimizer.update_fn(
        delta_params, state.server_opt_state)
    params = self._server_optimizer.apply_updates(state.params, updates)
    return AgnosticFedAvgState(
        params=params,
        server_opt_state=server_opt_state,
        domain_weights=domain_weights)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""EMNIST data loader."""

import collections
from typing import Optional, Tuple

from fedjax import core
import tensorflow as tf
import tensorflow_federated as tff

INIT_DOMAIN_WEIGHTS = (0.147, 0.853)  # HIGH_SCHOOL:CENSUS_FIELD.


def domain_id_fn(client_id: str) -> int:
  """Returns domain id for client id."""
  # These domain ids are based on NIST data source. For more details, see
  # https://s3.amazonaws.com/nist-srd/SD19/sd19_users_guide_edition_2.pdf.
  cid = int(client_id.split('_')[0][1:])
  if cid >= 2100 and cid <= 2599:
    return 0  # HIGH_SCHOOL.
  return 1  # CENSUS_FIELD.


def flip_and_expand(dataset: tf.data.Dataset) -> tf.data.Dataset:
  """Preprocesses dataset by expanding and inverting input pixel values.

  Args:
    dataset: Original source dataset.

  Returns:
    Dataset of ordered mapping of examples of the following structure:
      x: A numpy array of shape [28, 28, 1].
      y: A numpy array of shape [].
  """

  def map_fn(element):
    return collections.OrderedDict(
        x=tf.expand_dims(1.0 - element['pixels'], axis=-1),
        y=element['label'],
    )

  return dataset.map(map_fn, num_parallel_calls=tf.data.experimental.AUTOTUNE)


def load_data(
    only_digits: bool = True,
    cache_dir: Optional[str] = None
) -> Tuple[core.FederatedData, core.FederatedData]:
  """Loads EMNIST federated data with preprocessing.

  Args:
    only_digits: A bool whether to only include digits or not.
    cache_dir: Optional path to cache directory. If provided, files are
      downloaded over network to the specified cache directory.

  Returns:
    A tuple of federated data for train and test.
  """
  train_source, test_source = tff.simulation.datasets.emnist.load_data(
      only_digits=only_digits, cache_dir=cache_dir)
  return (train_source.preprocess(flip_and_expand),
          test_source.preprocess(flip_and_expand))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.client_trainer."""

import os

from absl.testing import flagsaver
from absl.testing import parameterized

from fedjax.core import client_trainer
from fedjax.core import dataset_util
from fedjax.core import evaluation_util
from fedjax.core import optimizer
from fedjax.core import test_util

import jax
from jax.lib import xla_bridge
import jax.numpy as jnp
import tensorflow as tf


def setUpModule():
  """Run all tests with 8 CPU devices."""
  global prev_xla_flags  # pylint: disable=global-variable-undefined
  prev_xla_flags = os.getenv('XLA_FLAGS')
  flags_str = prev_xla_flags or ''
  # Don't override user-specified device count, or other XLA flags.
  if 'xla_force_host_platform_device_count' not in flags_str:
    os.environ['XLA_FLAGS'] = (
        flags_str + ' --xla_force_host_platform_device_count=8')
  # Clear any cached backends so new CPU backend will pick up the env var.
  xla_bridge.get_backend.cache_clear()


def tearDownModule():
  """Reset to previous configuration in case other test modules will be run."""
  if prev_xla_flags is None:
    del os.environ['XLA_FLAGS']
  else:
    os.environ['XLA_FLAGS'] = prev_xla_flags
  xla_bridge.get_backend.cache_clear()


class DefaultClientTrainerTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._federated_algorithm = test_util.MockFederatedAlgorithm(
        num_clients=5, num_examples=20)
    self._federated_data = self._federated_algorithm.federated_data
    self._client_data_hparams = dataset_util.ClientDataHParams(batch_size=3)
    self._client_dataset = self._federated_data.create_tf_dataset_for_client(
        self._federated_data.client_ids[0]).batch(10)
    self._trainer = client_trainer.DefaultClientTrainer(
        self._federated_algorithm.model,
        optimizer.get_optimizer(
            optimizer.OptimizerName.SGD, learning_rate=0.01))

  def init_state(self):
    algo_state = self._federated_algorithm.init_state()
    return self._trainer.init_state(algo_state.params)

  def test_one_step(self):
    state = self.init_state()
    batch = next(self._client_dataset.as_numpy_iterator())
    rng = next(self._federated_algorithm._rng_seq)

    prev_metrics = self._federated_algorithm.model.evaluate(state.params, batch)
    state = self._trainer.one_step(state, batch, rng)
    metrics = self._federated_algorithm.model.evaluate(state.params, batch)

    self.assertEqual(state.num_examples, 10)
    self.assertLess(metrics['loss'].result(), prev_metrics['loss'].result())

  def test_loop(self):
    state = self.init_state()
    examples = zip(self._client_dataset.as_numpy_iterator(),
                   self._federated_algorithm._rng_seq)

    prev_metrics = evaluation_util.evaluate_single_client(
        self._client_dataset, self._federated_algorithm.model, state.params)
    state = self._trainer.loop(state, examples)
    metrics = evaluation_util.evaluate_single_client(
        self._client_dataset, self._federated_algorithm.model, state.params)

    self.assertEqual(state.num_examples, 20)
    self.assertLess(metrics['loss'], prev_metrics['loss'])

  def test_train_single_client_tf_dataset(self):
    state = client_trainer.train_single_client(
        dataset=self._client_dataset,
        client_trainer=self._trainer,
        init_client_trainer_state=self.init_state(),
        rng_seq=self._federated_algorithm._rng_seq)

    self.assertEqual(state.num_examples, 20)

  def test_train_single_client_iterator(self):
    state = client_trainer.train_single_client(
        dataset=self._client_dataset.as_numpy_iterator(),
        client_trainer=self._trainer,
        init_client_trainer_state=self.init_state(),
        rng_seq=self._federated_algorithm._rng_seq)

    self.assertEqual(state.num_examples, 20)

  @parameterized.named_parameters(
      {
          'testcase_name': 'sequential',
          'disable_parallel': 'true',
          'expected_num_examples': 20,
      }, {
          'testcase_name': 'parallel',
          'disable_parallel': 'false',
          'expected_num_examples': 18,  # (20 // 3) * 3
      }, {
          'testcase_name': 'auto',
          'disable_parallel': 'auto',
          'expected_num_examples': 18,  # (20 // 3) * 3
      })
  def test_train_multiple_clients(self, disable_parallel,
                                  expected_num_examples):
    with flagsaver.flagsaver(
        fedjax_experimental_disable_parallel=disable_parallel):
      state = self.init_state()
      states = client_trainer.train_multiple_clients(
          federated_data=self._federated_data,
          client_ids=self._federated_data.client_ids,
          client_trainer=self._trainer,
          init_client_trainer_state=state,
          rng_seq=self._federated_algorithm._rng_seq,
          client_data_hparams=self._client_data_hparams)
      states = list(states)

      self.assertLen(states, 5)
      for s in states:
        self.assertEqual(s.num_examples, expected_num_examples)


class ControlVariateTrainerTest(tf.test.TestCase, parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._federated_algorithm = test_util.MockFederatedAlgorithm(
        num_clients=5, num_examples=20)
    self._federated_data = self._federated_algorithm.federated_data
    self._client_data_hparams = dataset_util.ClientDataHParams(batch_size=3)
    self._client_dataset = self._federated_data.create_tf_dataset_for_client(
        self._federated_data.client_ids[0]).batch(10)
    self._trainer = client_trainer.ControlVariateTrainer(
        self._federated_algorithm.model,
        optimizer.get_optimizer(
            optimizer.OptimizerName.MOMENTUM, learning_rate=0.2, momentum=0.9))

  def init_state(self):
    algo_state = self._federated_algorithm.init_state()
    opt_state = self._trainer._base_optimizer.init_fn(algo_state.params)
    control_variate = jax.tree_map(jnp.ones_like, algo_state.params)
    return self._trainer.init_state(algo_state.params, opt_state,
                                    control_variate)

  def test_one_step(self):
    init_state = self.init_state()
    batch = next(self._client_dataset.as_numpy_iterator())
    rng = next(self._federated_algorithm._rng_seq)

    state = self._trainer.one_step(init_state, batch, rng)

    self.assertEqual(state.num_examples, 10)
    jax.tree_multimap(self.assertAllEqual, state.control_variate,
                      init_state.control_variate)
    jax.tree_multimap(self.assertNotAllEqual, state.params, init_state.params)

  def test_loop(self):
    init_state = self.init_state()
    examples = zip(self._client_dataset.as_numpy_iterator(),
                   self._federated_algorithm._rng_seq)

    state = self._trainer.loop(init_state, examples)

    self.assertEqual(state.num_examples, 20)
    jax.tree_multimap(self.assertAllEqual, state.control_variate,
                      init_state.control_variate)
    jax.tree_multimap(self.assertNotAllEqual, state.params, init_state.params)

  @parameterized.named_parameters(
      {
          'testcase_name': 'sequential',
          'disable_parallel': 'true',
          'expected_num_examples': 20,
      }, {
          'testcase_name': 'parallel',
          'disable_parallel': 'false',
          'expected_num_examples': 18,  # (20 // 3) * 3
      }, {
          'testcase_name': 'auto',
          'disable_parallel': 'auto',
          'expected_num_examples': 18,  # (20 // 3) * 3
      })
  def test_train_multiple_clients(self, disable_parallel,
                                  expected_num_examples):
    with flagsaver.flagsaver(
        fedjax_experimental_disable_parallel=disable_parallel):
      init_state = self.init_state()
      states = client_trainer.train_multiple_clients(
          federated_data=self._federated_data,
          client_ids=self._federated_data.client_ids,
          client_trainer=self._trainer,
          init_client_trainer_state=init_state,
          rng_seq=self._federated_algorithm._rng_seq,
          client_data_hparams=self._client_data_hparams)
      states = list(states)

      self.assertLen(states, 5)
      for s in states:
        self.assertEqual(s.num_examples, expected_num_examples)
        jax.tree_multimap(self.assertAllEqual, s.control_variate,
                          init_state.control_variate)


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.experimental.sqlite_federated_data."""

import itertools
import os.path
import zlib

from absl import flags
from absl.testing import absltest
from fedjax.experimental import serialization
from fedjax.experimental import sqlite_federated_data
import numpy as np
import numpy.testing as npt
import sqlite3

FLAGS = flags.FLAGS


class SQLiteFederatedDataTest(absltest.TestCase):

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    path = os.path.join(FLAGS.test_tmpdir, 'test_sqlite_federated_data.sqlite')

    # Create database file. First make sure the database file is empty.
    with open(path, 'w'):
      pass
    connection = sqlite3.connect(path)
    with connection:
      connection.execute("""
      CREATE TABLE federated_data (
        client_id BLOB NOT NULL PRIMARY KEY,
        data BLOB NOT NULL,
        num_examples INTEGER NOT NULL
      );""")
      for i in range(100):
        client_id = f'{i:04x}'.encode('ascii')
        features = {'x': np.arange(i + 1)}
        data = zlib.compress(serialization.msgpack_serialize(features))
        num_examples = i + 1
        connection.execute('INSERT INTO federated_data VALUES (?, ?, ?);',
                           [client_id, data, num_examples])
    connection.close()

    cls.FEDERATED_DATA = sqlite_federated_data.SQLiteFederatedData.new(path)

  def test_metadata(self):
    federated_data = self.FEDERATED_DATA
    self.assertEqual(federated_data.num_clients(), 100)
    self.assertCountEqual(federated_data.client_ids(),
                          [f'{i:04x}'.encode('ascii') for i in range(100)])
    self.assertCountEqual(
        federated_data.client_sizes(),
        [(f'{i:04x}'.encode('ascii'), i + 1) for i in range(100)])
    self.assertEqual(federated_data.client_size(b'0000'), 1)
    with self.assertRaises(KeyError):
      federated_data.client_size(b'xxxx')

  def _assert_correct_client_dataset(self,
                                     client_id,
                                     client_dataset,
                                     feature='x',
                                     preprocessor=None):
    self.assertLen(client_id, 4)
    self.assertRegex(client_id, br'00[0-f][0-f]')
    i = int(client_id, base=16)
    self.assertCountEqual(client_dataset.examples, [feature])
    npt.assert_array_equal(client_dataset.examples[feature], np.arange(i + 1))
    if preprocessor is None:
      self.assertIs(client_dataset.preprocessor,
                    self.FEDERATED_DATA._preprocess_batch)
    else:
      self.assertIs(client_dataset.preprocessor, preprocessor)

  def test_clients(self):
    client_ids = set()
    for client_id, client_dataset in self.FEDERATED_DATA.clients():
      self._assert_correct_client_dataset(client_id, client_dataset)
      client_ids.add(client_id)
    self.assertLen(client_ids, 100)

  def test_shuffled_clients(self):
    iterator = self.FEDERATED_DATA.shuffled_clients(buffer_size=10, seed=1)
    # Check client ids are shuffled.
    client_ids = []
    for client_id, client_dataset in itertools.islice(iterator, 10):
      self._assert_correct_client_dataset(client_id, client_dataset)
      client_ids.append(client_id)
    self.assertNotEqual(client_ids, [
        client_id
        for client_id, _ in itertools.islice(self.FEDERATED_DATA.clients(), 10)
    ])
    # Check iteration is repeated.
    for client_id, _ in itertools.islice(iterator, 190):
      client_ids.append(client_id)
    self.assertLen(client_ids, 200)
    self.assertNotEqual(client_ids[:100], client_ids[100:])

  def test_get_clients(self):
    client_ids = [b'0000', b'0010']
    for i, (client_id, client_dataset) in enumerate(
        self.FEDERATED_DATA.get_clients(client_ids)):
      self.assertEqual(client_id, client_ids[i])
      self._assert_correct_client_dataset(client_id, client_dataset)

    with self.assertRaises(KeyError):
      list(self.FEDERATED_DATA.get_clients([b'xxxx']))

  def test_get_client(self):
    self._assert_correct_client_dataset(b'0010',
                                        self.FEDERATED_DATA.get_client(b'0010'))
    with self.assertRaises(KeyError):
      self.FEDERATED_DATA.get_client(b'xxxx')

  def test_preprocess_client(self):
    federated_data = self.FEDERATED_DATA.preprocess_client(
        lambda x: {'y': x['x']})
    self.assertIs(federated_data._connection, self.FEDERATED_DATA._connection)
    self.assertIs(federated_data._parse_examples,
                  self.FEDERATED_DATA._parse_examples)
    self.assertIs(federated_data._start, self.FEDERATED_DATA._start)
    self.assertIs(federated_data._stop, self.FEDERATED_DATA._stop)
    self.assertIs(federated_data._preprocess_batch,
                  self.FEDERATED_DATA._preprocess_batch)
    self._assert_correct_client_dataset(
        *next(federated_data.clients()), feature='y')
    self._assert_correct_client_dataset(
        *next(federated_data.shuffled_clients(buffer_size=10, seed=1)),
        feature='y')
    self._assert_correct_client_dataset(
        *next(federated_data.get_clients([b'0000'])), feature='y')
    self._assert_correct_client_dataset(
        b'0010', federated_data.get_client(b'0010'), feature='y')

  def test_preprocess_batch(self):
    federated_data = self.FEDERATED_DATA.preprocess_batch(
        lambda x: {'z': x['x']})
    self.assertIs(federated_data._connection, self.FEDERATED_DATA._connection)
    self.assertIs(federated_data._parse_examples,
                  self.FEDERATED_DATA._parse_examples)
    self.assertIs(federated_data._start, self.FEDERATED_DATA._start)
    self.assertIs(federated_data._stop, self.FEDERATED_DATA._stop)
    self.assertIs(federated_data._preprocess_client,
                  self.FEDERATED_DATA._preprocess_client)
    self.assertCountEqual(
        federated_data._preprocess_batch({'x': np.arange(2)}), ['z'])
    self._assert_correct_client_dataset(
        *next(federated_data.clients()),
        preprocessor=federated_data._preprocess_batch)
    self._assert_correct_client_dataset(
        *next(federated_data.shuffled_clients(buffer_size=10, seed=1)),
        preprocessor=federated_data._preprocess_batch)
    self._assert_correct_client_dataset(
        *next(federated_data.get_clients([b'0000'])),
        preprocessor=federated_data._preprocess_batch)
    self._assert_correct_client_dataset(
        b'0010',
        federated_data.get_client(b'0010'),
        preprocessor=federated_data._preprocess_batch)

  def test_slice_metadata(self):
    federated_data = self.FEDERATED_DATA.slice(b'0001', b'0002')
    self.assertEqual(federated_data.num_clients(), 1)
    self.assertCountEqual(federated_data.client_ids(), [b'0001'])
    self.assertCountEqual(federated_data.client_sizes(), [(b'0001', 2)])
    self.assertEqual(federated_data.client_size(b'0001'), 2)
    for client_id in [b'0000', b'0002', b'xxxx']:
      with self.assertRaises(KeyError, msg=f'client_id: {client_id}'):
        federated_data.client_size(b'0000')

  def test_slice_iteration(self):
    federated_data = self.FEDERATED_DATA.slice(b'0001', b'0002')

    with self.subTest('clients'):
      clients = list(federated_data.clients())
      self.assertLen(clients, 1)
      self._assert_correct_client_dataset(*clients[0])

    with self.subTest('shuffled_clients'):
      shuffled_clients = list(
          itertools.islice(federated_data.shuffled_clients(buffer_size=10), 10))
      self.assertLen(shuffled_clients, 10)
      for client_id, client_dataset in shuffled_clients:
        self.assertEqual(client_id, b'0001')
        self._assert_correct_client_dataset(client_id, client_dataset)

    with self.subTest('get_clients'):
      clients = list(federated_data.get_clients([b'0001']))
      self.assertLen(clients, 1)
      self._assert_correct_client_dataset(*clients[0])
      with self.assertRaises(KeyError):
        list(federated_data.get_clients([b'0000']))

    with self.subTest('get_client'):
      self._assert_correct_client_dataset(b'0001',
                                          federated_data.get_client(b'0001'))
      with self.assertRaises(KeyError):
        federated_data.get_client(b'0000')


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.experimental.model."""

from absl.testing import absltest

from fedjax.experimental import metrics
from fedjax.experimental import model

import haiku as hk
import jax
from jax.experimental import stax
import numpy as np

train_loss = lambda b, p: metrics.unreduced_cross_entropy_loss(b['y'], p)
eval_metrics = {'accuracy': metrics.Accuracy()}


class ModelTest(absltest.TestCase):

  def check_model(self, model_):
    with self.subTest('init'):
      params = model_.init(jax.random.PRNGKey(0))
      num_params = sum(l.size for l in jax.tree_util.tree_leaves(params))
      self.assertEqual(num_params, 30)

    with self.subTest('apply_for_train'):
      batch = {
          'x': np.array([[1, 2], [3, 4], [5, 6]]),
          'y': np.array([7, 8, 9])
      }
      preds = model_.apply_for_train(params, batch, jax.random.PRNGKey(0))
      self.assertTupleEqual(preds.shape, (3, 10))

    with self.subTest('apply_for_eval'):
      preds = model_.apply_for_eval(params, batch)
      self.assertTupleEqual(preds.shape, (3, 10))

    with self.subTest('train_loss'):
      preds = model_.apply_for_train(params, batch, jax.random.PRNGKey(0))
      loss = model_.train_loss(batch, preds)
      self.assertTupleEqual(loss.shape, (3,))

  def test_create_model_from_haiku(self):

    def forward_pass(batch):
      return hk.Linear(10)(batch['x'])

    haiku_model = model.create_model_from_haiku(
        transformed_forward_pass=hk.transform(forward_pass),
        sample_batch={'x': np.ones((1, 2))},
        train_loss=train_loss,
        eval_metrics=eval_metrics)
    self.check_model(haiku_model)

  def test_create_model_from_stax(self):
    stax_init, stax_apply = stax.serial(stax.Dense(10))
    stax_model = model.create_model_from_stax(
        stax_init=stax_init,
        stax_apply=stax_apply,
        sample_shape=(-1, 2),
        train_loss=train_loss,
        eval_metrics=eval_metrics)
    self.check_model(stax_model)

  def test_evaluate_model(self):
    # Mock out Model.
    model_ = model.Model.new(
        init=lambda rng: None,  # Unused.
        apply_for_train=lambda params, batch, rng: None,  # Unused.
        apply_for_eval=lambda params, batch: batch.get('pred'),
        train_loss=lambda batch, preds: None,  # Unused.
        eval_metrics={
            'accuracy': metrics.Accuracy(),
            'loss': metrics.CrossEntropyLoss(),
        })
    params = None  # Unused.
    batches = [
        {
            'y': np.array([1, 0, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
        },
        {
            'y': np.array([0, 1, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
        },
        {
            'y': np.array([0, 1]),
            'pred': np.array([[1.2, 0.4], [2.3, 0.1]])
        },
    ]
    eval_results = model.evaluate_model(model_, params, batches)
    self.assertEqual(eval_results['accuracy'], 0.625)  # 5 / 8.
    self.assertAlmostEqual(eval_results['loss'], 0.8419596)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Transformations and utilities for specifying work for each client.

Many of these utilities work with and require PyTrees as input and output.
This is required when working with the underlying JAX transformations like
`jax.pmap` and `jax.jit`. For more information on what PyTrees are, refer to
https://jax.readthedocs.io/en/latest/pytrees.html.
"""

from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from fedjax.experimental.typing import BatchExample

import jax

PyTree = Any

# Server state that is passed to the client init to initialize the client state.
# This will contain aggregate information like the global model parameters that
# are then used as the starting point for client state.
ServerState = PyTree
# Optional persistent client state that can be passed alongside server state to
# the client init. For stateful client federated algorithms (usually in the
# cross-silo setting), this is how per client information will be passed along.
PersistentClientState = PyTree
# Intermittent client state passed as input and output for each client step.
# This will typically contain model parameters and optimizer state that are
# updated at each client step.
ClientState = PyTree
# Step results can be used to record any metrics over the course of running the
# for_each_client loop. For example, it can be used to record train metrics for
# each client like train loss and gradient norm.
StepResult = PyTree
# Final output from the for_each_client loop. This is usually a subset of the
# client state. However, more meaningful transformations can be done like adding
# noise to the final output for differential privacy.
ClientOutput = PyTree

ClientInit = Union[Callable[[ServerState], ClientState],
                   Callable[[ServerState, PersistentClientState], ClientState]]
ClientStep = Callable[[ClientState, BatchExample], Tuple[ClientState,
                                                         StepResult]]
ClientFinal = Callable[[ClientState], ClientOutput]

ClientId = str
# TODO(wuke): Replace this with the actual client dataset type.
ClientDatas = Iterable[Tuple[ClientId, Any]]
ForEachClient = Callable[
    [ClientDatas, ServerState, Optional[Dict[ClientId, PersistentClientState]]],
    Iterable[Tuple[ClientId, ClientOutput, List[StepResult]]]]


def for_each_client_jit(client_init: ClientInit, client_step: ClientStep,
                        client_final: ClientFinal) -> ForEachClient:
  """Creates a for each client function backed by `jax.jit`."""
  client_init_jit = jax.jit(client_init)
  client_step_jit = jax.jit(client_step)
  client_final_jit = jax.jit(client_final)

  def run(client_datas, server_state, persistent_client_states=None):
    for client_id, client_data in client_datas:
      step_results = []
      # pytype: disable=wrong-arg-count
      if persistent_client_states is None:
        client_state = client_init_jit(server_state)
      else:
        persistent_client_state = persistent_client_states[client_id]
        client_state = client_init_jit(server_state, persistent_client_state)
      # pytype: enable=wrong-arg-count
      for batch in client_data:
        client_state, step_result = client_step_jit(client_state, batch)
        step_results.append(step_result)
      client_output = client_final_jit(client_state)
      yield client_id, client_output, step_results

  return run


def for_each_client(client_init: ClientInit,
                    client_step: ClientStep,
                    client_final: ClientFinal = lambda s: s) -> ForEachClient:
  """Creates a function which maps over client datasets.

  For example, `for_each_client` could be used to define how to run client
  updates for each client in a federated training round. Another common use case
  of `for_each_client` is to run evaluation per client for a given set of model
  parameters.

  The underlying backend for `for_each_client` can differ depending on the
  available devices. For example, if multiple devies are available (e.g. TPU),
  `for_each_client` will use `jax.pmap` to parallelize across devices. It's also
  possible to manually specify which backend to use (helpful for debugging).

  The expected usage of `for_each_client` is as follows:

  ```python
  # Map over clients and count how many points are greater than `limit` for
  # each client. In addition to the total `count`, we'll also keep track of the
  # `num` per step in our step results.

  def client_init(server_state):
    return {'limit': server_state['limit'], 'count': 0.}

  def client_step(client_state, batch):
    limit = client_state['limit']
    num = jnp.sum(batch['x'] > limit)
    client_state = {'limit': limit, 'count': client_state['count'] + num}
    step_result = {'num': num}
    return client_state, step_result

  def client_final(client_state):
    return client_state['count']

  # Three clients with different data (`client_datasets`)
  # and starting counts (`client_infos`).
  client_datasets = [
      ('cid0', [{'x': jnp.array([1, 2, 3, 4])}, {'x': jnp.array([1, 2, 3])}]),
      ('cid1', [{'x': jnp.array([1, 2])}, {'x': jnp.array([1, 2, 3, 4, 5])}]),
      ('cid2', [{'x': jnp.array([1])}]),
  ]
  server_state = {'limit': jnp.array(2)}

  func = fedjax.experimental.for_each_client.for_each_client(
      client_init, client_step, client_final)
  print(list(func(client_datasets, server_state)))
  # [
  #   ('cid0', 3, [{'num': 2}, {'num': 1}]),
  #   ('cid1', 3, [{'num': 0}, {'num': 3}]),
  #   ('cid2', 0, [{'num': 0}]),
  # ]
  ```

  The same example with persistent client states:

  ```python
  def client_init_with_persistent_state(server_state, persistent_client_state):
    return {
        'limit': server_state['limit'],
        'count': persistent_client_state['count'],
    }

  persistent_client_states = {
      'cid0': {'count': jnp.array(2)},
      'cid1': {'count': jnp.array(0)},
      'cid2': {'count': jnp.array(1)},
  }

  func = fedjax.experimental.for_each_client.for_each_client(
      client_init_with_persistent_state, client_step, client_final)
  print(list(func(client_datasets, server_state, persistent_client_states)))
  # [
  #   ('cid0', 5, [{'num': 2}, {'num': 1}]),
  #   ('cid1', 3, [{'num': 0}, {'num': 3}]),
  #   ('cid2', 1, [{'num': 0}]),
  # ]
  ```

  Args:
    client_init: Function that initializes the internal client state from the
      starting server state as well as an optional persistent client state.
      The server state contains aggregate information like the global model
      parameters that are then used as the starting point to initialize the
      client state.
      The optional persistent client state is per client information that is
      expected to be persisted over multiple calls by the caller.
      The initialized internal client state is fed as input and output from
      `client_step` and `client_final`. Client state usually contains things
      like the model parameters and optimizer state that are updated at each
      `client_step`.
      This will be run once for each client.
    client_step: Function that takes the internal client state and a batch of
      examples as input and outputs a (possibly updated) client state along with
      any per step results. Per step results are usually diagnostics like train
      loss or gradient norm.
      This will be run for each batch for each client.
    client_final: Function that applies the final transformation on the internal
      client state to the desired, final client output. More meaningful
      transformations can be done here, like model update clipping.
      Defaults to just returning client state.
      This will be run once for each client.

  Returns:
    A for each client function that takes the client datas to map over, server
      state, and optional persistent client states as input and returns the
      outputs per client as specified in `client_final` along with any
      per client per step results.
  """
  for_each_client_backend = for_each_client_jit
  return for_each_client_backend(client_init, client_step, client_final)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Preprocessing and batching operations over client datasets.

#   Column based representation

The examples in a client dataset can be viewed as a table, where the rows are
the individual examples, and the columns are the features (labels are viewed as
a feature in this context).

We use a column based representation when loading a dataset into memory.
-   Each column is a NumPy array `x` of rank at least 1, where `x[i, ...]` is
the value of this feature for the `i`-th example.
-   The complete set of examples is a dict-like object, from `str` feature
names, to the corresponding column values.

Traditionally, a row based representation is used for representing the entire
dataset, and a column based representation is used for a single batch. In the
context of federated learning, an individual client dataset is small enough to
easily fit into memory so the same representation is used for the entire dataset
and a batch.

#   Preprocessor

Preprocessing on a batch of examples can be easily done via a chain of
functions. A `Preprocessor` object holds the chain of functions, and applies the
transformation on a batch of examples.

#   ClientDataset: examples + preprocessor

A ClientDataset is simply some examples in the column based representation,
accompanied by a Preprocessor. Its `batch()` method produces batches of examples
in a sequential order, suitable for evaluation. Its `shuffle_repeat_batch()`
method adds shuffling and repeating, making it suitable for training.
"""

from typing import Callable, Iterable, Iterator, Mapping, Optional

import numpy as np

# The same column based representation for examples in the entire client
# dataset, or in a single batch.
Examples = Mapping[str, np.ndarray]


def assert_consistent_rows(examples: Examples):
  """Asserts `examples` have consistent row sizes (the number of examples)."""
  sizes = {k: v.shape[0] for k, v in examples.items()}
  if not sizes:
    raise ValueError('No features in examples')
  it = iter(sizes.items())
  name, size = next(it)
  for k, v in it:
    if v != size:
      raise ValueError(
          f'Feature {name} has {size} rows, but feature {k} has {v} rows')


def num_examples(examples: Examples, validate: bool = True) -> int:
  if validate:
    assert_consistent_rows(examples)
  return len(next(iter(examples.values())))


def slice_examples(examples: Examples, index: slice) -> Examples:
  return {k: v[index] for k, v in examples.items()}


def pad_examples(examples: Examples, size: int, mask_key: str) -> Examples:
  """Pad examples to a fixed number of rows with 0 values.

  Args:
    examples: Examples to pad.
    size: The desired number of rows (i.e. the leading dimension size).
    mask_key: Name of the new feature storing whether each row is a non-padding
      example.

  Returns:
    Padded examples.

  Raises:
    ValueError: invalid inputs.
  """
  if mask_key in examples:
    raise ValueError(
        f'mask_key {mask_key!r} is already present in examples {examples}')
  current_size = num_examples(examples)
  if current_size == size:
    return examples
  elif current_size > size:
    raise ValueError(f'Cannot pad {current_size} examples to size {size}')
  result = {mask_key: np.arange(size) < current_size}
  for k, v in examples.items():
    padded = np.zeros((size,) + v.shape[1:], v.dtype)
    padded[:current_size] = v
    result[k] = padded
  return result


class Preprocessor:
  """A chain of preprocessing functions.

  `Preprocessor` holds a chain of preprocessing functions, and applies them
  in order on batched examples. Each individual preprocessing function operates
  over multiple examples, instead of just 1 example. For example,

  ```
  preprocessor = Preprocessor([
    # Flattens `pixels`.
    lambda x: {**x, 'pixels': x['pixels'].reshape([-1, 28 * 28])},
    # Introduce `binary_label`.
    lambda x: {**x, 'binary_label': x['label'] % 2},
  ])
  fake_emnist = {
    'pixels': np.random.uniform(size=(10, 28, 28)),
    'label': np.random.randint(10, size=(10,))
  }
  preprocessor(fake_emnist)
  # Produces a dict of [10, 28*28] "pixels", [10,] "label" and "binary_label".
  ```

  Given a `Preprocessor`, a new `Preprocessor` can be created with an additional
  preprocessing function appended to the chain,

  ```
  # Continuing from the previous example.
  new_preprocessor = preprocessor.append(
    lambda x: {**x, 'sum_pixels': np.sum(x['pixels'], axis=1)})
  new_preprocessor(fake_emnist)
  # Produces a dict of [10, 28*28] "pixels", [10,] "sum_pixels", "label" and
  # "binary_label".
  ```

  `Preprocessor` can process either the entire dataset, or a batch, because of
  the identical representation. Certain preprocessing can be done either at the
  dataset level, or at the batch level.

  ### Examples of preprocessing possible at either the dataset level, or the
  batch level

  Such preprocessing is deterministic, and strictly per-example.

  -   Casting a feature from `int8` to `float32`.
  -   Adding a new feature derived from existing features.
  -   Remove a feature (although the better place to do so is at the dataset
  level).

  A simple rule for deciding where to carry out the preprocessing in this case
  is the following,
  -   Does this make batching cheaper (e.g. removing features)? If so, do it at
  the dataset level.
  -   Otherwise, do it at the batch level.

  Assuming preprocessing time is linear in the number of examples, preprocessing
  at the batch level has the benefit of evenly distributing host compute work,
  which may overlap better with asynchronous JAX compute work on GPU/TPU.

  ### Examples of preprocessing only possible at the batch level

  -   Data augmentation (e.g. random cropping).
  -   Padding at the batch size dimension.

  ### Examples of preprocessing only possible at the dataset level
  -   Capping the number of examples.
  -   Altering what it means to be an example: e.g. in certain language model
  setups, sentences are concatenated and then split into equal sized chunks.
  """

  def __init__(self, fns: Iterable[Callable[[Examples], Examples]] = ()):
    self._fns = tuple(fns)

  def __call__(self, examples: Examples) -> Examples:
    if not self._fns:
      return examples
    # Make a copy to guard against fns that modify their input.
    out = dict(examples)
    for f in self._fns:
      out = f(out)
    assert_consistent_rows(out)
    return out

  def append(self, fn: Callable[[Examples], Examples]) -> 'Preprocessor':
    """Creates a new Preprocessor with fn added to the end."""
    return Preprocessor(self._fns + (fn,))

  def __str__(self) -> str:
    return f'Preprocessor({self._fns})'

  def __repr__(self) -> str:
    return str(self)


# A common default preprocessor that does nothing.
NoOpPreprocessor = Preprocessor()


class ClientDataset:
  """In memory client dataset backed by numpy ndarrays.

  Custom preprocessing on batches can be added via a preprocessor.

  This is only intended for efficient access to small datasets that fit in
  memory.
  """

  def __init__(self,
               examples: Examples,
               preprocessor: Preprocessor = NoOpPreprocessor):
    assert_consistent_rows(examples)
    self.examples = examples
    self.preprocessor = preprocessor

  def __len__(self) -> int:
    """Returns the number of examples in this dataset."""
    return num_examples(self.examples, validate=False)

  def __getitem__(self, index: slice) -> 'ClientDataset':
    """Returns a new ClientDataset with sliced examples."""
    if not isinstance(index, slice):
      raise ValueError(f'Only slicing is supported, got index {index!r}')
    return ClientDataset(
        slice_examples(self.examples, index), self.preprocessor)

  def batch(self,
            batch_size: int,
            num_batch_size_buckets: Optional[int] = None,
            mask_key: str = 'mask') -> Iterable[Examples]:
    """Produces preprocessed batches in a fixed sequential order.

    When the number of examples in the dataset is not a multiple of
    `batch_size`, the final batch may be smaller than `batch_size`.
    This may lead to a large number of JIT recompilations. This can be
    circumvented by padding the final batch to a small number of fixed sizes by
    specifying `num_batch_size_buckets`. If the final batch is padded, a new
    bool feature named `mask_key` is added so that each non-padding example is
    marked with True.

    We repeatedly halve the batch size up to `num_batch_size_buckets-1` times,
    until we find the smallest one that is also >= the size of the final batch.
    Therefore if `batch_size < 2^num_batch_size_buckets`, fewer bucket sizes
    will be actually used.

    Args:
      batch_size: Desired batch size.
      num_batch_size_buckets: Optional number of batch size buckets for the
        final batch.
      mask_key: The name of the new mask feature.

    Returns:
      An iterable object that can be iterated over multiple times.
    """
    return BatchView(self, batch_size, num_batch_size_buckets, mask_key)

  def shuffle_repeat_batch(self,
                           batch_size: int,
                           num_epochs: Optional[int] = None,
                           num_steps: Optional[int] = None,
                           seed: Optional[int] = None) -> Iterable[Examples]:
    """Produces preprocessed batches in a shuffled and repeated order.

    Shuffling is done without replacement, therefore for a dataset of N
    examples, the first `ceil(N/batch_size)` batches are guarranteed to cover
    the entire dataset.

    The number of batches produced from the iteration can be controlled by the
    `(num_epochs, num_steps)` iteration:
    -   If both are None, the shuffle-repeat process continues forever.
    -   If only `num_epochs` is set, as few batches as needed to go over the
    dataset this many passes are produced.
    -   If only `num_steps` is set, exactly this many batches are produced.
    -   If both `num_epochs` and `num_steps` are set, the fewer number of
    batches between the two conditions are produced.

    If reproducible iteration order is desired, a fixed `seed` can be used. When
    `seed` is None, repeated iteration over the same object may produce batches
    in a different order.

    Unlike `batch()`, batches from `shuffle_repeat_batch()` always contain
    exactly `batch_size` examples.

    Args:
      batch_size: The desired batch size.
      num_epochs: Optional number of passes to iterate over the dataset.
      num_steps: Optional number of batches to produce.
      seed: Optional random number generator seed.

    Returns:
      An iterable object that can be iterated over multiple times.
    """
    return ShuffleRepeatBatchView(
        self,
        batch_size=batch_size,
        num_epochs=num_epochs,
        num_steps=num_steps,
        seed=seed)


class BatchView:
  """View of ordered batches of a ClientDataset.

  See ClientDataset.batch() for the expected behavior.
  """

  def __init__(self, client_dataset: ClientDataset, batch_size: int,
               num_batch_size_buckets: Optional[int], mask_key: str):
    self._client_dataset = client_dataset
    self._data_size = len(client_dataset)
    self._batch_size = batch_size
    if num_batch_size_buckets is None:
      self._final_batch_size = None
    else:
      self._final_batch_size = _pick_final_batch_size(self._data_size,
                                                      self._batch_size,
                                                      num_batch_size_buckets)
    self._mask_key = mask_key

  def __iter__(self) -> Iterator[Examples]:
    for start in range(0, self._data_size, self._batch_size):
      stop = start + self._batch_size
      sliced = slice_examples(self._client_dataset.examples, slice(start, stop))
      processed = self._client_dataset.preprocessor(sliced)
      if stop <= self._data_size or self._final_batch_size is None:
        yield processed
      else:
        yield pad_examples(processed, self._final_batch_size, self._mask_key)


def _pick_final_batch_size(data_size: int, batch_size: int,
                           num_batch_size_buckets: int) -> int:
  """Picks the final batch size for a given dataset size."""
  # Determine the batch size for the final batch.
  final_batch_size = data_size % batch_size
  if final_batch_size == 0:
    # No padding necessary.
    return batch_size
  # final_batch_size in [1, batch_size)
  high, low, n = batch_size, batch_size // 2, 1
  # Find low < final_batch_size <= high
  while low >= final_batch_size and n < num_batch_size_buckets:
    high, low, n = low, low // 2, n + 1
  return high


class ShuffleRepeatBatchView:
  """View of shuffled and repeated batches of a ClientDataset.

  See ClientDataset.shuffle_repeat_batch() for the expected behavior.
  """

  def __init__(self, client_dataset: ClientDataset, batch_size: int,
               num_epochs: Optional[int], num_steps: Optional[int],
               seed: Optional[int]):
    self._client_dataset = client_dataset
    self._data_size = len(client_dataset)
    self._batch_size = batch_size
    if num_epochs is not None:
      self._num_steps = ((self._data_size * num_epochs + batch_size - 1) //
                         batch_size)
      if num_steps is not None:
        self._num_steps = min(num_steps, self._num_steps)
    elif num_steps is not None:
      self._num_steps = num_steps
    else:
      self._num_steps = None
    self._seed = seed

  def __iter__(self) -> Iterator[Examples]:
    buf = np.arange(self._data_size, dtype=np.int32)
    buf_size = buf.shape[0]
    # Start of unused portion of buf. We start with no unused values because we
    # haven't shuffled yet.
    i = buf_size
    num_steps = 0
    desired_num_steps = self._num_steps
    rng = np.random.RandomState(self._seed)
    while desired_num_steps is None or num_steps < desired_num_steps:
      # Find example indices of next batch.
      indices = np.zeros((self._batch_size,), dtype=np.int32)
      desired_size = indices.size
      filled = 0
      while filled < desired_size:
        available = buf_size - i
        if available == 0:
          rng.shuffle(buf)
          i = 0
          available = buf_size
        used = min(available, desired_size - filled)
        indices[filled:filled + used] = buf[i:i + used]
        i += used
        filled += used
      # Produce next batch.
      sliced = {k: v[indices] for k, v in self._client_dataset.examples.items()}
      yield self._client_dataset.preprocessor(sliced)
      num_steps += 1



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Training federated EMNIST model via federated averaging.

The model is a CNN with dropout, the client optimizer is vanilla SGD, and the
server optimizer is SGD with momentum.
"""

import collections

from absl import app
from absl import flags
from absl import logging

import fedjax
import simple_fed_avg
import jax
import tensorflow as tf

FLAGS = flags.FLAGS
# FederatedExperimentConfig flags.
flags.DEFINE_string('root_dir', '/tmp/emnist_simple_fed_avg',
                    'Root directory for experiment outputs (e.g. metrics).')
flags.DEFINE_integer('num_rounds', 1, 'Number of federated of training rounds.')
flags.DEFINE_integer('num_clients_per_round', 20,
                     'Number of clients per training round.')
flags.DEFINE_integer('sample_client_random_seed', None,
                     'Radom seed used to fix client sample per round.')
flags.DEFINE_integer(
    'checkpoint_frequency', 0,
    'Checkpoint frequency in rounds. If <= 0, no checkpointing is done.')
flags.DEFINE_integer('num_checkpoints_to_keep', 1,
                     'Maximum number of checkpoints to keep.')
flags.DEFINE_integer(
    'eval_frequency', 0,
    'Evaluation frequency in rounds. If <= 0, no evaluation is done.')

flags.DEFINE_string(
    'cache_dir', None,
    'Cache directory. If specified, files will be downloaded to disk. If '
    'unspecified, files are read directly over network.'
)
flags.DEFINE_bool('only_digits', False, 'Whether to use only digits or not.')
flags.DEFINE_integer('seed', 0, 'Random seed.')
flags.DEFINE_integer('client_batch_size', 20, 'Client local batch size.')
flags.DEFINE_integer('client_num_epochs', 1, 'Client local number of epochs.')
flags.DEFINE_float('client_learning_rate', 0.02,
                   'Client optimizer learning rate.')
flags.DEFINE_float('server_learning_rate', 3.0,
                   'Server optimizer learning rate.')
flags.DEFINE_float('server_momentum', 0.9, 'Server optimizer momentum.')
flags.DEFINE_integer('eval_batch_size', 512, 'Evaluation batch size.')


def main(_):
  logging.info('Device count: %d, host count: %d, local device count: %d',
               jax.device_count(), jax.host_count(), jax.local_device_count())
  logging.info('Devices: %s', jax.devices())
  fedjax.training.set_tf_cpu_only()

  only_digits = FLAGS.only_digits
  train_federated_data, test_federated_data = fedjax.datasets.emnist.load_data(
      only_digits=only_digits, cache_dir=FLAGS.cache_dir)
  # Set tf.data.Dataset preprocessing functions to the tff.simulation.ClientData
  # that applies batching according to evaluation hyperparameters.
  train_federated_data_for_eval = train_federated_data.preprocess(
      lambda ds: ds.batch(FLAGS.eval_batch_size))
  test_federated_data_for_eval = test_federated_data.preprocess(
      lambda ds: ds.batch(FLAGS.eval_batch_size))

  model = fedjax.models.emnist.create_conv_model(only_digits=only_digits)
  hparams = simple_fed_avg.SimpleFedAvgHParams(
      train_data_hparams=fedjax.ClientDataHParams(
          batch_size=FLAGS.client_batch_size,
          num_epochs=FLAGS.client_num_epochs,
          shuffle_buffer_size=100,
      ))
  rng_seq = fedjax.PRNGSequence(FLAGS.seed)

  config = fedjax.training.FederatedExperimentConfig(
      root_dir=FLAGS.root_dir,
      num_rounds=FLAGS.num_rounds,
      num_clients_per_round=FLAGS.num_clients_per_round,
      sample_client_random_seed=FLAGS.sample_client_random_seed,
      checkpoint_frequency=FLAGS.checkpoint_frequency,
      num_checkpoints_to_keep=FLAGS.num_checkpoints_to_keep,
      eval_frequency=FLAGS.eval_frequency,
  )
  tf.io.gfile.makedirs(config.root_dir)

  federated_averaging = simple_fed_avg.SimpleFedAvg(
      federated_data=train_federated_data,
      model=model,
      client_optimizer=fedjax.get_optimizer(
          fedjax.OptimizerName.SGD, learning_rate=FLAGS.client_learning_rate),
      server_optimizer=fedjax.get_optimizer(
          fedjax.OptimizerName.MOMENTUM,
          learning_rate=FLAGS.server_learning_rate,
          momentum=FLAGS.server_momentum),
      hparams=hparams,
      rng_seq=rng_seq,
  )

  periodic_eval_fn_map = collections.OrderedDict(
      fed_train_eval=fedjax.training.ClientEvaluationFn(
          train_federated_data_for_eval, model, config),
      fed_test_eval=fedjax.training.ClientEvaluationFn(
          test_federated_data_for_eval, model, config),
      full_test_eval=fedjax.training.FullEvaluationFn(
          test_federated_data_for_eval, model))
  final_eval_fn_map = collections.OrderedDict(
      full_test_eval=fedjax.training.FullEvaluationFn(
          test_federated_data_for_eval, model))
  fedjax.training.run_federated_experiment(
      config=config,
      federated_algorithm=federated_averaging,
      periodic_eval_fn_map=periodic_eval_fn_map,
      final_eval_fn_map=final_eval_fn_map)


if __name__ == '__main__':
  app.run(main)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.core.dataset_util."""

import collections

from absl.testing import parameterized
from fedjax.core import dataset_util
import numpy as np
import tensorflow as tf
import tensorflow_federated as tff


class DataTest(tf.test.TestCase, parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name='shuffle',
          hparams=dataset_util.ClientDataHParams(
              batch_size=3, num_epochs=1, shuffle_buffer_size=1),
          expected_num_batches=4),
      dict(
          testcase_name='num_epochs',
          hparams=dataset_util.ClientDataHParams(batch_size=3, num_epochs=5),
          expected_num_batches=17),
      dict(
          testcase_name='drop_remainder',
          hparams=dataset_util.ClientDataHParams(
              batch_size=3, num_epochs=2, drop_remainder=True),
          expected_num_batches=6),
      dict(
          testcase_name='keep_remainder',
          hparams=dataset_util.ClientDataHParams(
              batch_size=3, num_epochs=2, drop_remainder=False),
          expected_num_batches=7),
      dict(
          testcase_name='num_batches',
          hparams=dataset_util.ClientDataHParams(
              batch_size=3, num_epochs=2, num_batches=5),
          expected_num_batches=5),
  )
  def test_preprocess_tf_dataset(self, hparams, expected_num_batches):
    x = np.arange(10 * 2).reshape((10, 2))
    numpy_data = collections.OrderedDict(x=x, y=x, z=x)
    dataset = tf.data.Dataset.from_tensor_slices(numpy_data)

    batches = list(dataset_util.preprocess_tf_dataset(dataset, hparams))

    self.assertLen(batches, expected_num_batches)

  @parameterized.named_parameters(
      dict(
          testcase_name='select_all',
          all_client_ids=None,
          select_client_ids=None,
          expected_data=[
              collections.OrderedDict(x=[1, 2], y=[2, 4]),
              collections.OrderedDict(x=[3, 4], y=[6, 8]),
              collections.OrderedDict(x=[5, 6], y=[10, 12]),
              collections.OrderedDict(x=[7, 8], y=[14, 16]),
              collections.OrderedDict(x=[9, 10], y=[18, 20]),
          ]),
      dict(
          testcase_name='select_subset',
          all_client_ids=None,
          select_client_ids=['a', 'c'],
          expected_data=[
              collections.OrderedDict(x=[1, 2], y=[2, 4]),
              collections.OrderedDict(x=[3, 4], y=[6, 8]),
              collections.OrderedDict(x=[9, 10], y=[18, 20]),
          ]))
  def test_create_tf_dataset_for_clients(self, all_client_ids,
                                         select_client_ids, expected_data):
    tensor_slices_dict = collections.OrderedDict(
        a=collections.OrderedDict(x=[[1, 2], [3, 4]], y=[[2, 4], [6, 8]]),
        b=collections.OrderedDict(x=[[5, 6], [7, 8]], y=[[10, 12], [14, 16]]),
        c=collections.OrderedDict(x=[[9, 10]], y=[[18, 20]]),
    )
    mock_federated_data = tff.simulation.datasets.TestClientData(
        tensor_slices_dict)

    actual_dataset = dataset_util.create_tf_dataset_for_clients(
        mock_federated_data, select_client_ids)
    actual_data = tf.nest.map_structure(lambda x: x.numpy().tolist(),
                                        list(actual_dataset))

    actual_data = sorted(actual_data, key=lambda d: d['x'][0])
    expected_data = sorted(expected_data, key=lambda d: d['x'][0])
    self.assertSameStructure(actual_data, expected_data)

  def test_dataset_or_iterable(self):
    slices = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    # Access through TF dataset.
    for i, x in enumerate(
        dataset_util.iterate(tf.data.Dataset.from_tensor_slices(slices))):
      self.assertIsInstance(x, np.ndarray, msg=type(x))
      self.assertAllEqual(x, slices[i])
    # Access through direct iteration.
    for i, x in enumerate(dataset_util.iterate(slices)):
      self.assertIsInstance(x, np.ndarray, msg=type(x))
      self.assertAllEqual(x, slices[i])


if __name__ == '__main__':
  tf.test.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.experimental.metrics."""

from absl.testing import absltest
from absl.testing import parameterized

from fedjax.experimental import metrics

import jax.numpy as jnp
import numpy.testing as npt


class MeanStatTest(absltest.TestCase):

  def test_str(self):
    stat = metrics.MeanStat.new(2, 4)
    self.assertEqual(
        'MeanStat(accum=DeviceArray(2, dtype=int32), weight=DeviceArray(4, dtype=int32)) => 0.5',
        str(stat))

  def test_new(self):
    stat = metrics.MeanStat.new(jnp.array([2, 3, 1]), jnp.array([1, 0, 1]))
    npt.assert_array_equal(stat.accum, [2, 0, 1])
    npt.assert_array_equal(stat.weight, [1, 0, 1])

  def test_result(self):
    stat = metrics.MeanStat.new(2, 5)
    self.assertEqual(stat.result(), 0.4)

  def test_merge(self):
    stat_0 = metrics.MeanStat.new(1, 2)
    stat_1 = metrics.MeanStat.new(2, 3)
    merged_stat = stat_0.merge(stat_1)
    self.assertEqual(merged_stat.accum, 3)
    self.assertEqual(merged_stat.weight, 5)

  def test_reduce(self):
    stat = metrics.MeanStat.new(jnp.array([1, 2, 4]), jnp.array([1, 1, 0]))
    reduced_stat = stat.reduce()
    self.assertEqual(reduced_stat.accum, 3)
    self.assertEqual(reduced_stat.weight, 2)


class SumStatTest(absltest.TestCase):

  def test_str(self):
    stat = metrics.SumStat.new(2)
    self.assertEqual('SumStat(accum=DeviceArray(2, dtype=int32)) => 2',
                     str(stat))

  def test_result(self):
    stat = metrics.SumStat.new(2)
    self.assertEqual(stat.result(), 2)

  def test_merge(self):
    stat_0 = metrics.SumStat.new(1)
    stat_1 = metrics.SumStat.new(2)
    merged_stat = stat_0.merge(stat_1)
    self.assertEqual(merged_stat.accum, 3)

  def test_reduce(self):
    stat = metrics.SumStat.new(jnp.array([1, 2, 1]))
    reduced_stat = stat.reduce()
    self.assertEqual(reduced_stat.accum, 4)


class MetricsTest(parameterized.TestCase):

  def test_cross_entropy_loss(self):
    example = {'y': jnp.array(1)}
    prediction = jnp.array([1.2, 0.4])
    metric = metrics.CrossEntropyLoss()
    loss = metric.evaluate_example(example, prediction)
    self.assertAlmostEqual(loss.result(), 1.1711007)

  @parameterized.named_parameters(
      {
          'testcase_name': 'correct',
          'target': 2,
          'prediction': [0, 0, 1],
          'expected_result': 1.,
      }, {
          'testcase_name': 'incorrect',
          'target': 1,
          'prediction': [1, 0, 0],
          'expected_result': 0.,
      })
  def test_accuracy(self, target, prediction, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array(prediction)
    metric = metrics.Accuracy()
    accuracy = metric.evaluate_example(example, prediction)
    self.assertEqual(accuracy.result(), expected_result)

  def test_sequence_token_cross_entropy_loss(self):
    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = metrics.SequenceTokenCrossEntropyLoss()
    loss = metric.evaluate_example(example, prediction)
    self.assertAlmostEqual(loss.result(), 0.612331725)

  def test_sequence_cross_entropy_loss(self):
    example = {'y': jnp.array([1, 0, 1])}
    prediction = jnp.array([[1.2, 0.4], [2.3, 0.1], [0.3, 3.2]])
    metric = metrics.SequenceCrossEntropyLoss()
    loss = metric.evaluate_example(example, prediction)
    self.assertAlmostEqual(loss.result(), 1.2246635)

  def test_sequence_token_accuracy(self):
    example = {'y': jnp.array([1, 2, 2, 1, 0])}
    # prediction = [1, 0, 2, 1, 0].
    prediction = jnp.array([[0, 1, 0], [1, 0, 0], [0, 0, 1], [0, 1, 0],
                            [1, 0, 0]])
    metric = metrics.SequenceTokenAccuracy()
    accuracy = metric.evaluate_example(example, prediction)
    self.assertEqual(accuracy.result(), 0.75)  # 3 / 4.

  def test_sequence_token_count(self):
    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTokenCount(masked_target_values=(0, 2))
    count = metric.evaluate_example(example, prediction)
    self.assertEqual(count.result(), 3)

  @parameterized.named_parameters(
      {
          'testcase_name': 'untruncated',
          'target': [1, 2, 2, 3, 4, 0, 0],
          'expected_result': 0.,
      }, {
          'testcase_name': 'truncated',
          'target': [1, 2, 2, 3, 3, 3, 3],
          'expected_result': 1.,
      })
  def test_sequence_truncation_rate(self, target, expected_result):
    example = {'y': jnp.array(target)}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTruncationRate(eos_target_value=4)
    truncation_rate = metric.evaluate_example(example, prediction)
    self.assertEqual(truncation_rate.result(), expected_result)

  def test_sequence_token_oov_rate(self):
    example = {'y': jnp.array([1, 2, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceTokenOOVRate(oov_target_values=(2,))
    oov_rate = metric.evaluate_example(example, prediction)
    self.assertEqual(oov_rate.result(), 0.4)  # 2 / 5.

  def test_sequence_length(self):
    example = {'y': jnp.array([1, 2, 3, 4, 0, 0])}
    prediction = jnp.array([])  # Unused.
    metric = metrics.SequenceLength()
    sequence_length = metric.evaluate_example(example, prediction)
    self.assertEqual(sequence_length.result(), 4.0)


if __name__ == '__main__':
  absltest.main()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions and interfaces for single and multi client training."""

import abc
import functools
import itertools

from typing import Any, Generic, Iterable, Iterator, List, NamedTuple, Optional, Tuple, TypeVar
from absl import flags

from fedjax.core import dataset_util
from fedjax.core import prefetch
from fedjax.core import tree_util
from fedjax.core.model import Model
from fedjax.core.optimizer import Optimizer
from fedjax.core.typing import Batch
from fedjax.core.typing import FederatedData
from fedjax.core.typing import OptState
from fedjax.core.typing import Params
from fedjax.core.typing import PRNGKey
from fedjax.core.typing import PRNGSequence
from fedjax.core.typing import Updates
import jax
import jax.numpy as jnp
import jax.random as jrandom


T = TypeVar('T')

FLAGS = flags.FLAGS


flags.DEFINE_enum('fedjax_experimental_disable_parallel', 'true',
                  ['auto', 'true', 'false'],
                  'Set to `false` to parallelize `train_multiple_clients` on '
                  'multiple local devices via `jax.pmap`. '
                  'Defaults to `auto`, which will train in parallel only if '
                  'there is more than one local device available. '
                  'Training in parallel will automatically drop batches that '
                  'are not full batch size (the last batch).'
                  'Set to `true` to disable parallel and train sequentially, '
                  'meaning adding more devices does not help performance.')


class ClientTrainer(Generic[T], metaclass=abc.ABCMeta):
  """Interface for training a single client.

  This interface defines the core training components that interact with trainer
  state. Here, trainer state is a simple container holding only the necessary
  values for training e.g. model parameters and optimizer state.

  NOTE: Trainer state is distinct from the federated algorithm state. However,
  trainer state is typically derived from the federated algorithm state.
  """

  @abc.abstractmethod
  def init_state(self, *args, **kwargs) -> T:
    """Initializes the client trainer state."""

  @abc.abstractmethod
  def one_step(self, client_trainer_state: T, batch: Batch, rng: PRNGKey) -> T:
    """Runs one step of training on input batch and updates the trainer state.

    Args:
      client_trainer_state: Current client trainer state.
      batch: Batch of example numpy arrays of shape [client_batch_size, ...].
      rng: PRNGKey.

    Returns:
      Updated trainer state.
    """

  def loop(self, init_client_trainer_state: T,
           examples: Iterable[Tuple[Batch, PRNGKey]]) -> Any:
    """Runs training loop over multiple batches.

    For most usecases, this default implementation should work.

    Args:
      init_client_trainer_state: Initial client trainer state.
      examples: Sequence of batches and rngs, where batches are numpy arrays of
        shape [client_batch_size, ...].

    Returns:
      Output client trainer state that is typically just an updated version of
        the input `init_client_trainer_state`. However, output is flexible.
    """
    client_trainer_state = init_client_trainer_state
    for batch, rng in examples:
      client_trainer_state = self.one_step(client_trainer_state, batch, rng)
    return client_trainer_state


class DefaultClientTrainerState(NamedTuple):
  """State container for `DefaultClientTrainer`.

  Attributes:
    params: Pytree of model parameters.
    opt_state: Pytree of optimizer state.
    num_examples: Summed number of examples over all batches for a given client.
  """
  params: Params
  opt_state: OptState
  num_examples: float = 0.


class DefaultClientTrainer(ClientTrainer):
  """Default single client trainer."""

  def __init__(self, model: Model, optimizer: Optimizer):
    super().__init__()
    self._model = model
    self._optimizer = optimizer

  def init_state(self,
                 params: Params,
                 opt_state: Optional[OptState] = None,
                 num_examples: float = 0.) -> DefaultClientTrainerState:
    if opt_state is None:
      opt_state = self._optimizer.init_fn(params)
    return DefaultClientTrainerState(
        params=params, opt_state=opt_state, num_examples=num_examples)

  @functools.partial(jax.jit, static_argnums=0)
  def one_step(self, client_trainer_state: DefaultClientTrainerState,
               batch: Batch, rng: PRNGKey) -> DefaultClientTrainerState:
    backward_pass_output = self._model.backward_pass(
        client_trainer_state.params, batch, rng)
    params_updates, opt_state = self._optimizer.update_fn(
        backward_pass_output.grads, client_trainer_state.opt_state)
    params = self._optimizer.apply_updates(client_trainer_state.params,
                                           params_updates)
    return DefaultClientTrainerState(
        params=params,
        opt_state=opt_state,
        num_examples=client_trainer_state.num_examples +
        backward_pass_output.num_examples)


class ControlVariateTrainerState(NamedTuple):
  """State container for `ControlVariateTrainer`.

  Attributes:
    params: Pytree of model parameters.
    opt_state: Pytree of optimizer state.
    init_params: Pytree of initial model parameters.
    control_variate: Pytree of control variates that matches `params`.
    num_examples: Summed number of examples over all batches for a given client.
  """
  params: Params
  opt_state: OptState
  init_params: Params
  control_variate: Updates
  num_examples: float = 0.


class ControlVariateTrainer(ClientTrainer):
  """Trainer with control variates."""

  def __init__(self, model: Model, base_optimizer: Optimizer):
    super().__init__()
    self._model = model
    self._base_optimizer = base_optimizer

  def init_state(self,
                 params: Params,
                 opt_state: OptState,
                 control_variate: Updates,
                 num_examples: float = 0.) -> ControlVariateTrainerState:
    return ControlVariateTrainerState(
        params=params,
        opt_state=opt_state,
        init_params=params,
        control_variate=control_variate,
        num_examples=num_examples)

  @functools.partial(jax.jit, static_argnums=0)
  def one_step(self, client_trainer_state: ControlVariateTrainerState,
               batch: Batch, rng: PRNGKey) -> ControlVariateTrainerState:
    client_control_variate = self._model.backward_pass(
        client_trainer_state.init_params, batch, rng).grads
    backward_pass_output = self._model.backward_pass(
        client_trainer_state.params, batch, rng)
    grads = backward_pass_output.grads
    adjusted_grads = jax.tree_multimap(lambda g, cc, c: g - cc + c, grads,
                                       client_control_variate,
                                       client_trainer_state.control_variate)
    updates, _ = self._base_optimizer.update_fn(adjusted_grads,
                                                client_trainer_state.opt_state)
    params = self._base_optimizer.apply_updates(client_trainer_state.params,
                                                updates)
    return ControlVariateTrainerState(
        params=params,
        opt_state=client_trainer_state.opt_state,
        init_params=client_trainer_state.init_params,
        control_variate=client_trainer_state.control_variate,
        num_examples=client_trainer_state.num_examples +
        backward_pass_output.num_examples)


def train_single_client(dataset: dataset_util.DatasetOrIterable,
                        client_trainer: ClientTrainer[T],
                        init_client_trainer_state: T,
                        rng_seq: PRNGSequence) -> Any:
  """Trains model for a single client's dataset.

  Args:
    dataset: Dataset of examples [client_batch_size, d].
    client_trainer: ClientTrainer instance.
    init_client_trainer_state: Initial client trainer state. This will typically
      be derived from the federated algorithm state before calling
      `train_multiple_clients`.
    rng_seq: Random key generator.

  Returns:
    Output of client trainer that is typically just an updated version of the
      input `init_client_trainer_state`. However, output is flexible.
  """
  examples = zip(dataset_util.iterate(dataset), rng_seq)
  client_trainer_state = client_trainer.loop(init_client_trainer_state,
                                             examples)
  return client_trainer_state


def train_multiple_clients(
    federated_data: FederatedData, client_ids: List[str],
    client_trainer: ClientTrainer[T], init_client_trainer_state: T,
    rng_seq: PRNGSequence,
    client_data_hparams: dataset_util.ClientDataHParams) -> Iterator[Any]:
  """Trains separate model for each client and records client updates.

  Depending on the value of `--fedjax_experimental_disable_parallel`, this will
  be run either sequentially or in parallel across local devices via `jax.pmap`.

  Args:
    federated_data: Federated data separated per client.
    client_ids: Ids of clients to train.
    client_trainer: ClientTrainer instance.
    init_client_trainer_state: Initial client trainer state. This will typically
      be derived from algorithm state before calling `train_multiple_clients`.
    rng_seq: Random key generator.
    client_data_hparams: Hyperparameters for client dataset preparation.

  Yields:
    Output of client trainer that is typically just an updated version of the
      input `init_client_trainer_state`. However, output is flexible.
  """
  should_run_parallel = (
      FLAGS.fedjax_experimental_disable_parallel == 'false' or
      (FLAGS.fedjax_experimental_disable_parallel == 'auto' and
       jax.local_device_count() > 1))
  if should_run_parallel:
    yield from _train_multiple_clients_parallel(
        federated_data=federated_data,
        client_ids=client_ids,
        client_trainer=client_trainer,
        init_client_trainer_state=init_client_trainer_state,
        rng_seq=rng_seq,
        client_data_hparams=client_data_hparams)
    return

  preprocessed_federated_data = federated_data.preprocess(
      lambda ds: dataset_util.preprocess_tf_dataset(ds, client_data_hparams))
  for _, client_dataset in prefetch.PrefetchClientDatasetsIterator(
      preprocessed_federated_data, client_ids):
    examples = zip(dataset_util.iterate(client_dataset), rng_seq)
    client_trainer_state = client_trainer.loop(init_client_trainer_state,
                                               examples)
    yield client_trainer_state


def _get_fillvalue(data: List[List[Batch]]) -> Batch:
  """Creates a fill value that matches input batch pytree structure.

  Args:
    data: A list of list of batches, representing clients' batches.

  Returns:
    A batch of zeroes of the same structure as the input batch pytree.
  """
  for batch_list in data:
    if not batch_list:
      continue
    return jax.tree_map(jnp.zeros_like, batch_list[0])
  raise ValueError('All iterators fully consumed.')


def _one_step_f(trainer: ClientTrainer[T], client_trainer_state: T,
                batch: Batch, rng: PRNGKey) -> Any:
  """Runs one step of training and returns next trainer state and rng."""
  rng, next_rng = jrandom.split(rng)
  client_trainer_state = trainer.one_step(client_trainer_state, batch, rng)
  return client_trainer_state, next_rng


def _pmask(func, default_argnums=(0,)):
  """Wraps input function to add mask argument at first argument position.

  This is a convenience wrapper for masking functions that are intended to be
  run in parallel. Because most parallel functions in this module require that
  the leading axis dimension stay fixed, masking is necessary to support items
  that are less than the fixed size (i.e. remainder).

  NB: When using this function, remember that the argument indices will all be
  offset by +1 due to mask being set as the first argument.

  Args:
    func: Function that is intended to be run in parallel (e.g. train step).
    default_argnums: Argument index of default return value if mask is False.

  Returns:
    Wrapped mask function.
  """

  def masked_func(mask, *args, **kwargs):
    return jax.lax.cond(mask, lambda _: func(*args, **kwargs),
                        lambda _: tuple(args[i] for i in default_argnums), None)

  return masked_func

# Defaults are input trainer state and rng respectively.
_mask_step = _pmask(_one_step_f, default_argnums=(1, 3))
# static_broadcasted_argnums=1 points to the input ClientTrainer instance since
# pmask shifts all arguments forward by 1 position.
_pmap_step = jax.pmap(_mask_step, static_broadcasted_argnums=1)


def _train_multiple_clients_parallel(
    federated_data: FederatedData, client_ids: List[str],
    client_trainer: ClientTrainer[T], init_client_trainer_state: T,
    rng_seq: PRNGSequence,
    client_data_hparams: dataset_util.ClientDataHParams) -> Iterator[Any]:
  """Trains separate model for each client and records client updates.

  It has the same inputs and return values as `train_multiple_clients` above,
  but parallelizes the training across multiple devices if available.

  Args:
    federated_data: Federated data separated per client.
    client_ids: Ids of clients to train.
    client_trainer: ClientTrainer instance.
    init_client_trainer_state: Initial client trainer state. This will typically
      be derived from algorithm state before calling `train_multiple_clients`.
    rng_seq: Random key generator.
    client_data_hparams: Hyperparameters for client dataset preparation. The
      `drop_remainder` field is automatically set to True to ensure that all
      batches have the same batch size.

  Yields:
    Output of client trainer that is typically just an updated version of the
      input `init_client_trainer_state`. However, output is flexible.
  """
  client_data_hparams = client_data_hparams._replace(drop_remainder=True)
  client_data = []
  preprocessed_federated_data = federated_data.preprocess(
      lambda ds: dataset_util.preprocess_tf_dataset(ds, client_data_hparams))
  for _, client_dataset in prefetch.PrefetchClientDatasetsIterator(
      preprocessed_federated_data, client_ids):
    client_data.append(list(dataset_util.iterate(client_dataset)))
  # Sort by length to group similarly sized datasets together to minimize the
  # amount of fillvalues needed.
  client_data = sorted(client_data, key=len)
  # TODO(b/177346980): Handle case where all clients' data is empty by padding.
  fillvalue = _get_fillvalue(client_data)

  num_local_devices = jax.local_device_count()
  init_stack_state = tree_util.tree_broadcast(
      init_client_trainer_state, axis_size=num_local_devices)
  stack_rng = jrandom.split(next(rng_seq), num_local_devices)

  quotient, remainder = divmod(len(client_ids), num_local_devices)
  num_iterations = quotient + bool(remainder)
  for i in range(num_iterations):
    stack_state = init_stack_state
    streams = client_data[num_local_devices * i: num_local_devices * (i + 1)]
    # Handle number of clients not divisible by num_devices.
    if len(streams) < num_local_devices:
      client_count = len(streams)
      streams.extend(
          [[fillvalue] for _ in range(num_local_devices - client_count)])
    else:
      client_count = num_local_devices

    for batches in itertools.zip_longest(*streams, fillvalue=fillvalue):
      mask = jnp.array([b is not fillvalue for b in batches])
      stack_mask = tree_util.tree_stack(mask)
      stack_batch = tree_util.tree_stack(batches)
      # Mask must be the first input.
      stack_state, stack_rng = _pmap_step(stack_mask, client_trainer,
                                          stack_state, stack_batch, stack_rng)

    # Unstack stack_state, yield each one.
    for j, final_state in enumerate(tree_util.tree_unstack(stack_state)):
      if j < client_count:
        yield final_state



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fedjax.experimental.for_each_client."""

from absl.testing import absltest

from fedjax.experimental import for_each_client

import jax.numpy as jnp
import numpy as np

# Map over clients and count how many points are greater than `limit` for
# each client. In addition to the total `count`, we'll also keep track of the
# `num` per step in our step results.


def client_init(server_state):
  return {'limit': server_state['limit'], 'count': 0.}


def client_step(client_state, batch):
  limit = client_state['limit']
  num = jnp.sum(batch['x'] > limit)
  client_state = {'limit': limit, 'count': client_state['count'] + num}
  step_result = {'num': num}
  return client_state, step_result


def client_final(client_state):
  return client_state['count']


# Variant of the task above where we provide a starting count for each client
# as part of persistent_client_state.


def client_init_with_persistent_state(server_state, persistent_client_state):
  return {
      'limit': server_state['limit'],
      'count': persistent_client_state['count'],
  }


class ForEachClientTest(absltest.TestCase):

  def test_for_each_client_jit(self):
    # Three clients with different data (`client_datasets`)
    # and starting counts (`client_infos`).
    client_datasets = [
        ('cid0', [{'x': jnp.array([1, 2, 3, 4])}, {'x': jnp.array([1, 2, 3])}]),
        ('cid1', [{'x': jnp.array([1, 2])}, {'x': jnp.array([1, 2, 3, 4, 5])}]),
        ('cid2', [{'x': jnp.array([1])}]),
    ]
    server_state = {'limit': jnp.array(2)}

    with self.subTest('without_persistent_client_state'):
      func = for_each_client.for_each_client_jit(client_init, client_step,
                                                 client_final)
      results = list(func(client_datasets, server_state))
      np.testing.assert_equal(results, [
          ('cid0', jnp.array(3),
           [{'num': jnp.array(2)}, {'num': jnp.array(1)}]),
          ('cid1', jnp.array(3),
           [{'num': jnp.array(0)}, {'num': jnp.array(3)}]),
          ('cid2', jnp.array(0),
           [{'num': jnp.array(0)}]),
      ])

    with self.subTest('with_persistent_client_state'):
      persistent_client_states = {
          'cid0': {'count': jnp.array(2)},
          'cid1': {'count': jnp.array(0)},
          'cid2': {'count': jnp.array(1)},
      }
      func = for_each_client.for_each_client_jit(
          client_init_with_persistent_state, client_step, client_final)
      results = list(
          func(client_datasets, server_state, persistent_client_states))
      np.testing.assert_equal(results, [
          ('cid0', jnp.array(5),
           [{'num': jnp.array(2)}, {'num': jnp.array(1)}]),
          ('cid1', jnp.array(3),
           [{'num': jnp.array(0)}, {'num': jnp.array(3)}]),
          ('cid2', jnp.array(1),
           [{'num': jnp.array(0)}]),
      ])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A small library for working with evaluation metrics such as accuracy.

##  Quick overview

To evaluate model predictions, use a `Metric` object such as `Accuracy`. We
recommend `fedjax.experimental.model.evaluate_model()` in most scenarios, which
runs model prediction, and evaluation, on batches of N examples at a time for
greater computational efficiency.

```python
# Mock out Model.
model = fedjax.experimental.model.Model.new(
    init=lambda _: None,  # Unused.
    apply_for_train=lambda _, _, _: None,  # Unused.
    apply_for_eval=lambda _, batch: batch.get('pred'),
    train_loss=lambda _, _: None,  # Unused.
    eval_metrics{'accuracy': metrics.Accuracy()})
params = None  # Unused.
batches = [{'y': np.array([1, 0]), 'pred': np.array([[1.2, 0.4], [2.3, 0.1]])},
           {'y': np.array([1, 1]), 'pred': np.array([[0.3, 3.2], [2.1, 4.3]])}]
eval_results = fedjax.experimental.model.evaluate_model(model, params, batches)
print(eval_results)
# {'accuracy': 0.75}
```

A `Metric` object has 2 methods:

-   `zero()`: Returns an initial value for accumulating the statistic for this
metric.
-   `evaluate_example(example, prediction)`: Returns the statistic from
evaluating a single example, given the training `example` and the model
`prediction`.

Most `Metric`s follow the following convention for convenience:

-   `example` is a dict-like object from `str` to `jnp.ndarray`.
-   `prediction` is either a single `jnp.ndarray`, or a dict-like object from
`str` to `jnp.ndarray`.

Conceptually, we can also use a simple for loop to evaluate a collection of
examples and model predictions:

```python
# By default, the `Accuracy` metric treats `example['y']` as the true label, and
# `prediction` as a single `jnp.ndarray` of class scores.
metric = Accuracy()
stat = metric.zero()
# We are iterating over individual examples, not batches.
for example, prediction in [({'y': jnp.array(1)}, jnp.array([0., 1.])),
                            ({'y': jnp.array(0)}, jnp.array([1., 0.])),
                            ({'y': jnp.array(1)}, jnp.array([1., 0.])),
                            ({'y': jnp.array(0)}, jnp.array([2., 0.]))]:
  stat = stat.merge(metric.evaluate_example(example, prediction))
print(stat.result())
# 0.75
```

In practice, for greater computational efficiency, we run model prediction not
on a single example, but a batch of N examples at a time.
`fedjax.experimental.model.evaluate_model()` is provides a simple way to do so.
Under the hood, it calls `evaluate_batch()`.

```python
metric = Accuracy()
stat = metric.zero()
# We are iterating over batches.
for batch_example, batch_prediction in [
  ({'y': jnp.array([1, 0])}, jnp.array([[0., 1.], [1., 0.]])),
  ({'y': jnp.array([1, 0])}, jnp.array([[1., 0.], [2., 0.]]))]:
  stat = stat.merge(evaluate_batch(metric, batch_example, batch_prediction))
print(stat.result())
# 0.75
```

##  Under the hood

For most users, it is sufficient to know how to use existing `Metric` subclasses
such as `Accuracy` with `fedjax.experimental.model.evaluate_model()`. This
section is intended for those who would like to write new metrics.

### From algebraic structures to `Metric` and `Stat`

There are 2 abstraction in this library, `Metric` and `Stat`. Before going into
details of these classes, let's first consider a few abstract properties related
to evaluation metrics, using accuracy as an example.

When evaluating accuracy on a dataset, we wish to know the proportion of
examples that are correctly predicted by the model. Because a dataset might be
too large to fit into memory, we need to divide the work by partitioning the
dataset into smaller subsets, evaluate each separately, and finally somehow
combine the results. Assuming the subsets can be of different sizes, although
the accuracy value is a single number, we cannot just average the accuracy
values from each partition to arrive at the overall accuracy. Instead, we need 2
numbers from each subset:

-   The number of examples in this subset,
-   How many of them are correctly predicted.

We call these two numbers from each subset a _statistic_. The domain (the set of
possible values) of the statistic in the case of accuracy is `{(0, 0)} ∪ {(a,
b) | a >= 0, b > 0}`.

With the numbers of examples and correct predictions from 2 disjoint subsets, we
add the numbers up to get the number of examples and correct predictions for the
union of the 2 subsets. We call this operation from 2 statistics into 1 a
`merge` operation.

Let `f(S)` be the function that gives us the statistic from a subset of
examples. It is easy to see for two disjoint subsets `A` and `B`, `merge(f(A),
f(B))` should be equal to `f(A ∪ B)`. If no such `merge` exists, we cannot
evaluate the dataset by partitioning the work. This requirement alone implies
the domain of a statistic, and the `merge` operation forms a specific algebraic
structure (a commutative monoid).

-   `I := f(empty set)` is one and the only identity element w.r.t. `merge`
(i.e. `merge(I, x) == merge(x, I) == x`.
-   `merge()` is commutative and associative.

Further, we can see `f(S)` can be defined just knowing two types of values:

-   `f(empty set)`, i.e. `I`;
-   `f({x})` for any single example `x`.

For any other subset `S`, we can derive the value of `f(S)` using these values
and `merge`. `Metric` is simply the `f(S)` function above, defined in 2
corresponding parts:

-   `Metric.zero()` is `f(empty set)`.
-   `Metric.evaluate_example(example, prediction)` is `f({x})` for a single
example.

On the other hand, `Stat` stores a single statistic, a `merge` method for
combining two, and a `result` method for producing the final metric value.

To implement `Accuracy` as a subclass of `Metric`, we first need to know what
`Stat` to use. In this case, the statistic domain and `merge` is implemented by
a `MeanStat`. A `MeanStat` holds two values:

-   `accum` is the weighted sum of values, i.e. the number of correct
predictions in the case of accuracy.
-   `weight`is the sum of weights, i.e. the number of examples in the case of
accuracy.

`merge` adds up the respective `accum` and `weight` from two `MeanStat` objects.

Sometimes, a new `Stat` subclass is necessary. In that case, it is very
important to make sure the implementation has a clear definition of the domain,
and the `merge` operation adheres to the properties regarding identity element,
commutativity, and associativity (e.g. if we unknowingly allow pairs of `(x, 0)`
for `x != 0` into the domain of a `MeanStat`, `merge((x, 0), (a, b))` will
produce a statistic that leads to incorrect final metric values, i.e. `(a+x)/b`,
instead of `a/b`).

### Batching `Stat`s

In most cases, the final value of an evaluation is simply a scalar, and the
corresponding statistic is also a tuple of a few scalar values. However, for the
same reason why `jax.vmap` is a lot more efficient than a for loop, it is a lot
more efficient to store multiple `Stat` values as a `Stat` of arrays, instead of
a list of `Stat` objects. Thus instead of a list `[MeanStat(1, 2), MeanStat(3,
4), MeanStat(5, 6)]` (call these "rank 0" `Stat`s), we can batch the 3
statitstics as a single `MeanStat(jnp.array([1, 3, 5]), jnp.array([2, 4, 6]))`.
A `Stat` object holding a single statistic is a "rank 0" `Stat`. A `Stat` object
holding a vector of statistics is a "rank 1" `Stat`. Similarly, a `Stat` object
may also hold a matrix, a 3D array, etc, of statistics. These are higher rank
`Stat`s.

In the end, we want just 1 final metric value instead of a length 3 vector, or a
2x2 matrix, of metric values. To finally go back to a single statistic (), we
need to `merge` statistics stored in these arrays. Each `Stat` subclass provides
a `reduce` method to do just that. The combination of `jax.vmap` over
`Metric.evaluate_example`, and `Stat.reduce`, is how we get an efficient
`evaluate_batch()` function (of course, the real `evaluate_batch` is `jax.jit`'d
so that the same `jax.vmap` transformation etc does not need to happen over and
over.

Importantly, for a higher rank `Stat` (rank >= 1), only `reduce` needs to work.
Other operations such as `merge` and `result` only needs to support rank 0
`Stat`s (because we can get the batched version of these via `jax.vmap`).

```python
metric = Accuracy()
stat = metric.zero()
# We are iterating over batches.
for batch_example, batch_prediction in [
  ({'y': jnp.array([1, 0])}, jnp.array([[0., 1.], [1., 0.]])),
  ({'y': jnp.array([1, 0])}, jnp.array([[1., 0.], [2., 0.]]))]:
  # Get a batch of statistics as a single Stat object.
  batch_stat = jax.vmap(metric.evaluate_example)(batch_example,
  batch_prediction)
  # Merge the reduced single statistic onto the accumulator.
  stat = stat.merge(batch_stat.reduce())
print(stat.result())
# 0.75
```

Being able to batch `Stat`s also allow us to do other interesting things, for
example,
-   `evaluate_batch` accepts an optional per-example `mask` so it can work on
padded batches.
-   We can define a `PerDomain` metric for any base metric so that we can get
accuracy where examples are partitioned by a domain id.

### Creating a new metric

Most likely, a new metric will just return a `MeanStat` or a `SumStat`. If
that's the case, simply following implement a new `Metric` following the
guidelines in `Metric`'s class docstring.

If a new `Stat` is necessary, following the guidelines in `Stat`'s docstring.
"""

import abc
import functools
from typing import Optional, Tuple

from fedjax import core
from fedjax.experimental.typing import BatchExample
from fedjax.experimental.typing import BatchPrediction
from fedjax.experimental.typing import SingleExample
from fedjax.experimental.typing import SinglePrediction

import jax
import jax.numpy as jnp


class Stat(metaclass=abc.ABCMeta):
  """Stat keeps some statistic, along with operations over them.

  Most users will only need to interact with a `Stat` object via `result()`.

  For those who need to create new metrics, please first read the "Under the
  hood" section of the module docstring.

  Most `Stat`'s domain (the set of possible statistic values) has constraints,
  it is thus usually a good practice to offer and use factory methods to
  construct new `Stat` objects instead of directly assigning the fields.

  To work with various jax constructs, a concrete `Stat` should be a PyTree.
  This is easily achieved with `fedjax.dataclass`.

  A `Stat` may hold either a single statistic (a rank 0 `Stat`), or an array of
  statistics (a higher rank `Stat`). `result` and `merge` only needs to work on
  a rank 0 `Stat`. `reduce` only needs to work on a higher rank `Stat`.
  """

  @abc.abstractmethod
  def result(self) -> jnp.ndarray:
    """Calculates the metric value from the statistic value.

    For example, `MeanStat.result()` calculates a weighted average.

    Returns:
      The return value of `result()` must be a `jnp.ndarray`.
    """

  @abc.abstractmethod
  def merge(self, other: 'Stat') -> 'Stat':
    """Merges two `Stat` objects into a new `Stat` with merged statistics.

    Args:
      other: Another `Stat` object of the same type.

    Returns:
      A new `Stat` object of the same type with merged statistics.
    """

  @abc.abstractmethod
  def reduce(self, axis: Optional[int] = 0):
    """Reduces a higher rank statistic along a given `axis`.

    See the class docstring for details.

    Args:
      axis: An integer axis index, or `None`.

    Returns:
      A new `Stat` object of the same type.
    """

  def __str__(self) -> str:
    return f'{repr(self)} => {self.result()}'


@core.dataclass
class MeanStat(Stat):
  """Statistic for weighted mean calculation.

  A `MeanStat` maintains the weighted sum (`accum`) and the sum of weights
  (`weight`) for calculating the weighted mean.

  Prefer using the `new` or `single` factory methods instead of directly
  assigning to fields.
  """
  accum: jnp.ndarray
  weight: jnp.ndarray

  @classmethod
  def new(cls, accum, weight) -> 'MeanStat':
    """Creates a sanitized `MeanStat`.

    The domain of a weighted mean statistic is:

    ```
    {(0, 0)} ∪ {(a, b) | a >= 0, b > 0}
    ```

    `new()` sanitizes values outside the domain into the identity (zeros).

    Args:
      accum: A value convertible to `jnp.ndarray`.
      weight: A value convertible to `jnp.ndarray`.

    Returns:
      The sanitized `MeanStat`.
    """
    weight = jnp.maximum(0, jnp.array(weight, copy=False))
    accum = jnp.where(weight == 0, 0, jnp.array(accum, copy=False))
    return cls(accum, weight)

  def result(self) -> jnp.ndarray:
    return jnp.where(self.weight == 0, 0, self.accum / self.weight)

  def merge(self, other: 'MeanStat') -> 'MeanStat':
    accum = self.accum + other.accum
    weight = self.weight + other.weight
    return MeanStat.new(accum, weight)

  def reduce(self, axis: Optional[int] = 0):
    return MeanStat.new(
        jnp.sum(self.accum, axis=axis), jnp.sum(self.weight, axis=axis))


@core.dataclass
class SumStat(Stat):
  """Statistic for summing values.

  A `SumStat` simply maintains the sum of values (`accum`).
  """
  accum: jnp.ndarray

  @classmethod
  def new(cls, accum: jnp.ndarray) -> 'SumStat':
    return cls(jnp.array(accum, copy=False))

  def result(self) -> jnp.ndarray:
    return self.accum

  def merge(self, other: 'SumStat') -> 'SumStat':
    return SumStat.new(self.accum + other.accum)

  def reduce(self, axis: Optional[int] = 0):
    return SumStat.new(jnp.sum(self.accum, axis=axis))


class Metric(metaclass=abc.ABCMeta):
  """`Metric` is the conceptual metric (like accuracy).

  It defines two methods:

  -   `evaluate_example()` evaluates a single example, and returns a `Stat`
  object.
  -   `zero()` returns the identity value for what `evaluate_example()` returns.

  Given a `Metric` object `m`, let

  -   `u = m.zero()`
  -   `v = m.evaluate_example(...)`

  We require that

  -   `type(u) == type(v)`.
  -   `u.merge(v) == v.merge(u) == v`.
  -   Components of `u` has the same shape as the counter parts in `v`.
  """

  @abc.abstractmethod
  def zero(self) -> Stat:
    """Returns a `Stat` such that merging with it is an identity operation.

    e.g. for accuracy: `MeanStat.new(0., 0.)`

    Returns:
      `Stat` identity value.
    """

  @abc.abstractmethod
  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> Stat:
    """Evaluates a single example.

    e.g. for accuracy: `MeanStat.new(num_correct, num_total)`

    Args:
      example: A single input example (e.g. one sentence for language).
      prediction: Output for `example` from `Model.apply_for_eval`.

    Returns:
      `Stat` value.
    """


def apply_mask(mask: jnp.ndarray, a: jnp.ndarray,
               b: jnp.ndarray) -> jnp.ndarray:
  """Applies mask on the leading dimension."""
  rank = max(len(a.shape), len(b.shape))
  return jnp.where(jnp.expand_dims(mask, tuple(range(1, rank))), a, b)


@functools.partial(jax.jit, static_argnums=0)
def evaluate_batch(metric: Metric,
                   batch_example: BatchExample,
                   batch_prediction: BatchPrediction,
                   batch_mask: Optional[jnp.ndarray] = None):
  """Evaluates a batch using a metric."""
  batch_stat = jax.vmap(metric.evaluate_example)(batch_example,
                                                 batch_prediction)
  if batch_mask is not None:
    batch_stat = jax.tree_util.tree_multimap(
        functools.partial(apply_mask, batch_mask), batch_stat, metric.zero())
  return batch_stat.reduce()


def unreduced_cross_entropy_loss(targets: jnp.ndarray,
                                 preds: jnp.ndarray) -> jnp.ndarray:
  """Returns unreduced cross entropy loss."""
  num_classes = preds.shape[-1]
  log_preds = jax.nn.log_softmax(preds)
  one_hot_targets = jax.nn.one_hot(targets, num_classes)
  return -jnp.sum(one_hot_targets * log_preds, axis=-1)


@core.dataclass
class CrossEntropyLoss(Metric):
  """Metric for cross entropy loss.

  Attributes:
    target_key: Key name in `example` for target.
    pred_key: Key name in `prediction` for unnormalized model output pred.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes cross entropy loss for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for loss for a single example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    loss = unreduced_cross_entropy_loss(target, pred)
    return MeanStat.new(loss, 1.)


@core.dataclass
class Accuracy(Metric):
  """Metric for accuracy.

  Attributes:
    target_key: Key name in `example` for target.
    pred_key: Key name in `prediction` for unnormalized model output pred.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes accuracy for a single example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for accuracy for a single example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    correct = (target == jnp.argmax(pred, axis=-1)).astype(jnp.float32)
    return MeanStat.new(correct, 1.)


def _target_weight(
    target: jnp.ndarray, masked_target_values: Tuple[jnp.ndarray,
                                                     ...]) -> jnp.ndarray:
  target_weight = jnp.ones_like(target, dtype=jnp.float32)
  for mv in masked_target_values:
    target_weight *= (target != mv)
  return target_weight


@core.dataclass
class SequenceTokenCrossEntropyLoss(Metric):
  """Metric for token cross entropy loss for a sequence example.

  Attributes:
    target_key: Key name in `example` for target.
    pred_key: Key name in `prediction` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token cross entropy loss for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for token loss for a single sequence example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    target_weight = _target_weight(target, self.masked_target_values)
    token_loss = unreduced_cross_entropy_loss(target, pred)
    return MeanStat.new(
        jnp.sum(token_loss * target_weight), jnp.sum(target_weight))


@core.dataclass
class SequenceCrossEntropyLoss(Metric):
  """Metric for total cross entropy loss for a sequence example.

  Attributes:
    target_key: Key name in `example` for target.
    pred_key: Key name in `prediction` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes total cross entropy loss for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for total loss for a single sequence example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    target_weight = _target_weight(target, self.masked_target_values)
    token_loss = unreduced_cross_entropy_loss(target, pred)
    # Change weight from number of non masked target tokens to 1 if the sequence
    # contains any non masked tokens or 0 if the entire sequence is masked.
    return MeanStat.new(
        jnp.sum(token_loss * target_weight), jnp.any(target_weight))


@core.dataclass
class SequenceTokenAccuracy(Metric):
  """Metric for token accuracy for a sequence example.

  Attributes:
    target_key: Key name in `example` for target.
    pred_key: Key name in `prediction` for unnormalized model output pred.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  pred_key: Optional[str] = None
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token accuracy for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for token accuracy for a single sequence example.
    """
    target = example[self.target_key]
    pred = prediction if self.pred_key is None else prediction[self.pred_key]
    target_weight = _target_weight(target, self.masked_target_values)
    correct = (target == jnp.argmax(pred, axis=-1)).astype(jnp.float32)
    return MeanStat.new(
        jnp.sum(correct * target_weight), jnp.sum(target_weight))


@core.dataclass
class SequenceTokenCount(Metric):
  """Metric for count of non masked tokens for a sequence example.

  Attributes:
    target_key: Key name in `example` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> SumStat:
    return SumStat.new(0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> SumStat:
    """Computes total number of non masked tokens in a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `SumStat` for count of non masked tokens for a single sequence example.
    """
    del prediction
    target = example[self.target_key]
    target_weight = _target_weight(target, self.masked_target_values)
    return SumStat.new(jnp.sum(target_weight))


@core.dataclass
class SequenceTruncationRate(Metric):
  """Metric for truncation rate for a sequence example.

  Attributes:
    eos_target_value: Target value denoting end of sequence. Truncated sequences
      will not have this value.
    target_key: Key name in `example` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  eos_target_value: jnp.ndarray
  target_key: str = 'y'
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes truncation rate for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for truncation rate for a single sequence.
    """
    del prediction
    target = example[self.target_key]
    target_weight = _target_weight(target, self.masked_target_values)
    not_empty = jnp.sum(jnp.any(target_weight))
    target_is_truncated = jnp.all(target != self.eos_target_value)
    return MeanStat.new(target_is_truncated * not_empty, not_empty)


@core.dataclass
class SequenceTokenOOVRate(Metric):
  """Metric for out-of-vocabulary (OOV) rate for a sequence example.

  Attributes:
    oov_target_values: Target values denoting out-of-vocabulary values.
    target_key: Key name in `example` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  oov_target_values: Tuple[jnp.ndarray, ...]
  target_key: str = 'y'
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes token out of vocabulary rate for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for token out of vocabulary rate for a single sequence.
    """
    del prediction
    target = example[self.target_key]
    target_weight = _target_weight(target, self.masked_target_values)
    target_oov = jnp.ones_like(target, dtype=jnp.float32)
    for oov_value in self.oov_target_values:
      target_oov *= (target == oov_value)
    return MeanStat.new(
        jnp.sum(target_oov * target_weight), jnp.sum(target_weight))


@core.dataclass
class SequenceLength(Metric):
  """Metric for length for a sequence example.

  Attributes:
    target_key: Key name in `example` for target.
    masked_target_values: Target values that should be ignored in computation.
      This is typically used to ignore padding values in computation.
  """
  target_key: str = 'y'
  masked_target_values: Tuple[jnp.ndarray, ...] = (0,)

  def zero(self) -> MeanStat:
    return MeanStat.new(0., 0.)

  def evaluate_example(self, example: SingleExample,
                       prediction: SinglePrediction) -> MeanStat:
    """Computes non masked length for a single sequence example.

    Args:
      example: One example with target in range [0, num_classes) of shape [1].
      prediction: Unnormalized prediction for `example` of shape [num_classes].

    Returns:
      `MeanStat` for non masked length for a single sequence.
    """
    del prediction
    target = example[self.target_key]
    target_weight = _target_weight(target, self.masked_target_values)
    return MeanStat.new(jnp.sum(target_weight), jnp.sum(jnp.any(target_weight)))
